from builtins import str
from builtins import object
import gzip
import io
import os
import shutil
import sqlite3
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
import seaflowpy as sfp

# pylint: disable=redefined-outer-name

@pytest.fixture()
def evt_df():
    return sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00")["df"]


@pytest.fixture()
def params():
    # Params created with popcycle function
    # create.filter.params(740, 33759, 19543, 19440)
    # taking the 50.0 quantile values.
    return pd.DataFrame.from_dict({
        "width": [2500, 2500, 2500],
        "notch_small_D1": [0.614, 0.656, 0.698],
        "notch_small_D2": [0.651, 0.683, 0.714],
        "notch_large_D1": [1.183, 1.635, 2.087],
        "notch_large_D2": [1.208, 1.632, 2.056],
        "offset_small_D1": [1418, 0, -1418],
        "offset_small_D2": [1080, 0, -1047],
        "offset_large_D1": [-17791, -33050, -48309],
        "offset_large_D2": [-17724, -32038, -46352],
        "quantile": [2.5, 50.0, 97.5]
    })


@pytest.fixture()
def tmpout(tmpdir):
    """Setup to test complete filter workflow"""
    # Copy db with filtering params
    db_one = str(tmpdir.join("testcruise_one.db"))
    shutil.copyfile("tests/testcruise_paramsonly_one_param.db", db_one)
    os.chmod(db_one, 0o664)  # make the db writeable

    db_plan = str(tmpdir.join("testcruise_plan.db"))
    shutil.copyfile("tests/testcruise_paramsonly_plan.db", db_plan)
    os.chmod(db_plan, 0o664)  # make the db writeable

    evt_path = "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00"
    return {
        "db_one": db_one,
        "db_plan": db_plan,
        "oppdir": str(tmpdir.join("oppdir")),
        "tmpdir": str(tmpdir),
        "evt_df": sfp.fileio.read_evt_labview(str(evt_path))["df"],
        "evt_path": evt_path,
        "file_dates": pd.read_parquet("tests/file_dates.parquet")
    }


class TestOpenV1:
    @pytest.mark.benchmark(group="evt-read")
    def test_read_evt_valid(self, benchmark):
        data = benchmark(sfp.fileio.read_evt_labview, "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00")
        assert len(data["df"].index) == 40000
        assert data["version"] == "v1"
        assert list(data["df"]) == sfp.particleops.COLUMNS

    @pytest.mark.benchmark(group="evt-read")
    def test_read_evt_valid_gz(self, benchmark):
        data = benchmark(sfp.fileio.read_evt_labview, "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz")
        assert len(data["df"].index) == 40000
        assert data["version"] == "v1"
        assert list(data["df"]) == sfp.particleops.COLUMNS

    def test_read_evt_truncated_gz(self, tmpout):
        truncpath = os.path.join(tmpout["tmpdir"], "2014-07-04T00-03-02+00-00.gz")
        with open("tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz", "rb") as infh:
            with open(truncpath, "wb") as outfh:
                outfh.write(infh.read(400000))
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview(truncpath)

    def test_read_evt_empty(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-06-02+00-00")

    def test_read_evt_zero_header(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-09-02+00-00")

    def test_read_evt_short_header(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-12-02+00-00")

    def test_read_evt_more_data_than_header_count(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-21-02+00-00")

    def test_read_evt_less_data_than_header_count(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt/2014_185/2014-07-04T00-27-02+00-00")


class TestOpenV2:
    @pytest.mark.benchmark(group="evt-read")
    def test_read_evt_valid_v2(self, benchmark):
        data = benchmark(sfp.fileio.read_evt_labview, "tests/testcruise_evt_v2/2014_185/2014-07-04T00-00-02+00-00")
        assert len(data["df"].index) == 40000
        assert data["version"] == "v2"
        assert list(data["df"]) == sfp.particleops.COLUMNS2

    @pytest.mark.benchmark(group="evt-read")
    def test_read_evt_valid_gz_v2(self, benchmark):
        data = benchmark(sfp.fileio.read_evt_labview, "tests/testcruise_evt_v2/2014_185/2014-07-04T00-03-02+00-00.gz")
        assert len(data["df"].index) == 40000
        assert data["version"] == "v2"
        assert list(data["df"]) == sfp.particleops.COLUMNS2

    def test_read_evt_truncated_gz_v2(self, tmpout):
        truncpath = os.path.join(tmpout["tmpdir"], "2014-07-04T00-03-02+00-00.gz")
        with open("tests/testcruise_evt_v2/2014_185/2014-07-04T00-03-02+00-00.gz", "rb") as infh:
            with open(truncpath, "wb") as outfh:
                outfh.write(infh.read(100000))
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview(truncpath)

    def test_read_evt_empty_v2(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt_v2/2014_185/2014-07-04T00-06-02+00-00")

    def test_read_evt_zero_header_v2(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt_v2/2014_185/2014-07-04T00-09-02+00-00")

    def test_read_evt_short_header_v2(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt_v2/2014_185/2014-07-04T00-12-02+00-00")

    def test_read_evt_more_data_than_header_count_v2(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt_v2/2014_185/2014-07-04T00-21-02+00-00")

    def test_read_evt_less_data_than_header_count_v2(self):
        with pytest.raises(sfp.errors.FileError):
            _data = sfp.fileio.read_evt_labview("tests/testcruise_evt_v2/2014_185/2014-07-04T00-27-02+00-00")


class TestFilter:
    def test_mark_focused_no_params(self, evt_df):
        with pytest.raises(ValueError):
            _df = sfp.particleops.mark_focused(evt_df, None)

    def test_mark_focused_empty_params(self, evt_df):
        with pytest.raises(ValueError):
            _df = sfp.particleops.mark_focused(evt_df, pd.DataFrame.from_dict({}))

    @pytest.mark.benchmark(group="evt-filter")
    def test_mark_focused_with_set_params(self, evt_df, params, benchmark):
        orig_df = evt_df.copy()  # make a copy before any possible modifications

        # Should not modify original
        new_evt_df = benchmark(sfp.particleops.mark_focused, evt_df, params)
        assert not (new_evt_df is evt_df)  # returned a new dataframe
        assert orig_df.equals(evt_df)  # original dataframe is the unmodified
        assert len(new_evt_df.index) == 40000
        assert len(new_evt_df[new_evt_df["q2.5"]].index) == 423
        assert len(new_evt_df[new_evt_df["q50"]].index) == 107
        assert len(new_evt_df[new_evt_df["q97.5"]].index) == 85
        assert len(sfp.particleops.select_focused(new_evt_df).index) == 426

    def test_mark_focused_with_set_params_inplace(self, evt_df, params, benchmark):
        orig_df = evt_df.copy()  # make a copy before any possible modifications

        # Should modify original
        new_evt_df = benchmark(sfp.particleops.mark_focused, evt_df, params, inplace=True)
        assert new_evt_df is evt_df  # returned the same dataframe
        assert not orig_df.equals(evt_df)  # original dataframe is the modified
        assert len(new_evt_df.index) == 40000
        assert len(new_evt_df[new_evt_df["q2.5"]].index) == 423
        assert len(new_evt_df[new_evt_df["q50"]].index) == 107
        assert len(new_evt_df[new_evt_df["q97.5"]].index) == 85
        assert len(sfp.particleops.select_focused(new_evt_df).index) == 426

    def test_noise_filter(self, evt_df):
        """Events with zeroes in all of D1, D2, and fsc_small are noise"""
        # There are events which could be considered noise (no signal in any of
        # D1, D2, or fsc_small
        assert np.any((evt_df["D1"] == 0) & (evt_df["D2"] == 0) & (evt_df["fsc_small"] == 0)) == True

        noise = sfp.particleops.mark_noise(evt_df)
        assert noise.sum() == 72
        signal_df = evt_df[~noise]

        # Correct event count
        assert len(signal_df.index) == 39928

        # No events are all zeroes D1, D2, and fsc_small
        assert np.any((signal_df["D1"] == 0) & (signal_df["D2"] == 0) & (signal_df["fsc_small"] == 0)) == False

    def test_saturation_filter(self, evt_df):
        """Events with max D1 or max D2"""
        sat = sfp.particleops.mark_saturated(evt_df)
        assert sat.sum() == 211

        d1_max = evt_df["D1"] == evt_df["D1"].max()
        d2_max = evt_df["D2"] == evt_df["D2"].max()
        assert len(evt_df[~(d1_max | d2_max)]) == 39789

    def test_all_quantiles(self):
        df = pd.DataFrame(
            {
                "x": [True, True, True],
                "q2.5": [True, False, True],
                "q50": [True, True, True],
                "q97.5": [False, True, False]
            }
        )
        assert sfp.particleops.all_quantiles(df) == True

        df = pd.DataFrame(
            {
                "x": [True, True, True],
                "q2.5": [True, False, True],
                "q50": [True, True, True],
                "q97.5": [False, False, False]
            }
        )
        assert sfp.particleops.all_quantiles(df) == False


class TestTransform:
    def test_linearize_four_values(self):
        input_df = pd.DataFrame({
            "fsc_small": [32768, 65536],
            "D1": [0, 1]
        })
        output_df = pd.DataFrame({
            "fsc_small": [56.234132519, 3162.2776601684],
            "D1": [1, 1.0001229789]
        })
        npt.assert_array_almost_equal(
            sfp.particleops.linearize_particles(input_df, columns=["fsc_small", "D1"]),
            output_df
        )

    def test_linearize_copy(self, evt_df):
        orig_df = evt_df.copy()
        npt.assert_array_equal(orig_df, evt_df)
        t_df = sfp.particleops.linearize_particles(evt_df, columns=["fsc_small", "D1"])
        assert t_df is not evt_df
        with pytest.raises(AssertionError):
            npt.assert_array_equal(orig_df, t_df)

    def test_log_four_values(self):
        input_df = pd.DataFrame({
            "fsc_small": [56.234132519, 3162.2776601684],
            "D1": [1, 1.0001229789]
        })
        output_df = pd.DataFrame({
            "fsc_small": [32768, 65536],
            "D1": [0, 1]
        })
        npt.assert_array_almost_equal(
            sfp.particleops.log_particles(input_df, columns=["fsc_small", "D1"]),
            output_df
        )

    def test_log_copy(self, evt_df):
        evt_df = sfp.particleops.linearize_particles(evt_df, columns=["fsc_small", "D1"])
        orig_df = evt_df.copy()
        npt.assert_array_equal(orig_df, evt_df)
        t_df = sfp.particleops.log_particles(evt_df, columns=["fsc_small", "D1"])
        assert t_df is not evt_df
        with pytest.raises(AssertionError):
            npt.assert_array_equal(orig_df, t_df)


class TestOutput:
    def test_sqlite3_opp_counts_and_params(self, tmpout, params):
        sf_file = sfp.seaflowfile.SeaFlowFile(tmpout["evt_path"])
        df = tmpout["evt_df"]
        df = sfp.particleops.mark_focused(df, params, inplace=True)

        raw_count = len(df.index)
        signal_count = len(df[df["noise"] == False].index)

        vals = sfp.db.prep_opp(sf_file.file_id, df, raw_count, signal_count, "UUID")
        sfp.db.save_opp_to_db(vals, tmpout["db_one"])
        con = sqlite3.connect(tmpout["db_one"])
        sqlitedf = pd.read_sql_query("SELECT * FROM opp", con)

        try:
            opp_evt_ratio = len(df[df["q50"]].index) / len(df[df["noise"] == False].index)
        except ZeroDivisionError:
            opp_evt_ratio = 0.0

        assert sf_file.file_id == sqlitedf["file"][1]
        assert sqlitedf["filter_id"][1] == "UUID"
        assert sqlitedf["quantile"][1] == 50
        npt.assert_array_equal(
            [107, 39928, 40000, opp_evt_ratio],
            sqlitedf[["opp_count", "evt_count", "all_count", "opp_evt_ratio"]].values[1]
        )


    def test_sqlite3_opp_counts_and_params_empty(self, tmpout, params):
        sf_file = sfp.seaflowfile.SeaFlowFile(tmpout["evt_path"])
        df = sfp.particleops.empty_df()
        df = sfp.particleops.mark_focused(df, params, inplace=True)

        raw_count = len(df.index)
        signal_count = len(df[df["noise"] == False].index)

        vals = sfp.db.prep_opp(sf_file.file_id, df, raw_count, signal_count, "UUID")
        sfp.db.save_opp_to_db(vals, tmpout["db_one"])
        con = sqlite3.connect(tmpout["db_one"])
        sqlitedf = pd.read_sql_query("SELECT * FROM opp", con)

        assert sf_file.file_id == sqlitedf["file"][1]
        assert sqlitedf["filter_id"][1] == "UUID"
        assert sqlitedf["quantile"][1] == 50
        npt.assert_array_equal(
            [0, 0, 0, 0.0],
            sqlitedf[["opp_count", "evt_count", "all_count", "opp_evt_ratio"]].values[1]
        )

    def test_binary_evt_output(self, tmpout):
        sfile = sfp.seaflowfile.SeaFlowFile(tmpout["evt_path"])
        evtdir = os.path.join(tmpout["tmpdir"], "evtdir")
        evt_df = sfp.fileio.read_evt_labview(tmpout["evt_path"])["df"]

        # Output to binary uncompressed file
        sfp.fileio.write_evt_labview(evt_df, sfile.file_id, evtdir, gz=False)
        # Make sure EVT binary file written can be read back as EVT and
        # DataFrame is the the same
        out_evt_path = os.path.join(evtdir, sfile.file_id)
        reread_evt_df = sfp.fileio.read_evt_labview(out_evt_path)["df"]
        npt.assert_array_equal(evt_df, reread_evt_df)
        # Check that output evt binary file matches input file
        input_evt = io.open(tmpout["evt_path"], "rb").read()
        new_evt = io.open(out_evt_path, "rb").read()
        assert input_evt == new_evt

    def test_binary_evt_output_gz(self, tmpout):
        sfile = sfp.seaflowfile.SeaFlowFile(tmpout["evt_path"])
        evtdir = os.path.join(tmpout["tmpdir"], "evtdir")
        evt_df = sfp.fileio.read_evt_labview(tmpout["evt_path"])["df"]

        # Output to gzipped binary file
        sfp.fileio.write_evt_labview(evt_df, sfile.file_id, evtdir)
        # Make sure EVT binary file written can be read back as EVT and
        # DataFrame is the the same
        out_evt_path = os.path.join(evtdir, sfile.file_id + ".gz")
        reread_evt_df = sfp.fileio.read_evt_labview(out_evt_path)["df"]
        npt.assert_array_equal(evt_df, reread_evt_df)
        # Check that output evt binary file matches input file
        input_evt = io.open(tmpout["evt_path"], "rb").read()
        new_evt = gzip.open(out_evt_path).read()
        assert input_evt == new_evt


class TestMultiFileFilter(object):
    def test_multi_file_filter_local(self, tmpout):
        """Test multi-file filtering and ensure output can be read back OK"""
        # python setup.py test doesn't play nice with pytest and
        # multiprocessing, so we use one core here
        sfp.filterevt.filter_evt_files(
            tmpout["file_dates"],
            dbpath=tmpout["db_one"],
            opp_dir=str(tmpout["oppdir"]),
            worker_count=1
        )

        opp_dfs = [
            pd.read_parquet(os.path.join(tmpout["oppdir"], "2014-07-04T00-00-00+00-00.1H.opp.parquet")),
            pd.read_parquet(os.path.join(tmpout["oppdir"], "2014-07-04T01-00-00+00-00.1H.opp.parquet"))
        ]
        expected_opp_dfs = [
            pd.read_parquet("tests/testcruise_opp_one_param/2014-07-04T00-00-00+00-00.1H.opp.parquet"),
            pd.read_parquet("tests/testcruise_opp_one_param/2014-07-04T01-00-00+00-00.1H.opp.parquet")
        ]
        pdt.assert_frame_equal(opp_dfs[0], expected_opp_dfs[0], check_exact=False)
        pdt.assert_frame_equal(opp_dfs[1], expected_opp_dfs[1], check_exact=False)

        # Check numbers stored in opp table are correct
        opp_table = sfp.db.get_opp_table(tmpout["db_one"])
        expected_opp_table = sfp.db.get_opp_table("tests/testcruise_full_one_param.db")

        pdt.assert_frame_equal(opp_table, expected_opp_table, check_exact=False)

        # Check that outlier table has entry for every file
        outlier_table = sfp.db.get_outlier_table(tmpout["db_one"])
        expected_outlier_table = sfp.db.get_outlier_table("tests/testcruise_full_one_param.db")
        pdt.assert_frame_equal(outlier_table, expected_outlier_table)

    def test_multi_file_filter_local_v2(self, tmpout):
        """Test multi-file filtering on v2 data and ensure output can be read back OK"""
        # python setup.py test doesn't play nice with pytest and
        # multiprocessing, so we use one core here
        file_dates = tmpout["file_dates"].copy()
        file_dates["path"] = file_dates["path_v2"]
        sfp.filterevt.filter_evt_files(
            file_dates,
            dbpath=tmpout["db_plan"],
            opp_dir=str(tmpout["oppdir"]),
            worker_count=1
        )

        opp_dfs = [
            pd.read_parquet(os.path.join(tmpout["oppdir"], "2014-07-04T00-00-00+00-00.1H.opp.parquet")),
            pd.read_parquet(os.path.join(tmpout["oppdir"], "2014-07-04T01-00-00+00-00.1H.opp.parquet"))
        ]
        expected_opp_dfs = [
            pd.read_parquet("tests/testcruise_opp_plan/2014-07-04T00-00-00+00-00.1H.opp.parquet"),
            pd.read_parquet("tests/testcruise_opp_plan/2014-07-04T01-00-00+00-00.1H.opp.parquet")
        ]
        pdt.assert_frame_equal(opp_dfs[0], expected_opp_dfs[0], check_exact=False)
        pdt.assert_frame_equal(opp_dfs[1], expected_opp_dfs[1], check_exact=False)

        # Check numbers stored in opp table are correct
        opp_table = sfp.db.get_opp_table(tmpout["db_plan"])
        expected_opp_table = sfp.db.get_opp_table("tests/testcruise_full_plan.db")

        pdt.assert_frame_equal(opp_table, expected_opp_table, check_exact=False)

        # Check that outlier table has entry for every file
        outlier_table = sfp.db.get_outlier_table(tmpout["db_plan"])
        expected_outlier_table = sfp.db.get_outlier_table("tests/testcruise_full_plan.db")
        pdt.assert_frame_equal(outlier_table, expected_outlier_table)



from builtins import str
import datetime
import pkgutil
import sqlite3
import uuid
import pandas as pd
from . import errors
from . import particleops
from .seaflowfile import SeaFlowFile


def create_db(dbpath):
    """Create or complete database"""
    schema_text = pkgutil.get_data(__name__, 'data/popcycle.sql').decode('UTF-8', 'ignore')
    executescript(dbpath, schema_text)


def save_filter_params(dbpath, vals):
    create_db(dbpath)
    # NOTE: values inserted must be in the same order as fields in opp
    # table. Defining that order in a list here makes it easier to verify
    # that the right order is used.
    field_order = [
        "id",
        "date",
        "quantile",
        "beads_fsc_small",
        "beads_D1",
        "beads_D2",
        "width",
        "notch_small_D1",
        "notch_small_D2",
        "notch_large_D1",
        "notch_large_D2",
        "offset_small_D1",
        "offset_small_D2",
        "offset_large_D1",
        "offset_large_D2"
    ]
    # Construct values string with named placeholders
    values_str = ", ".join([":" + f for f in field_order])
    sql_insert = "INSERT OR REPLACE INTO filter VALUES ({})".format(values_str)
    id_ = str(uuid.uuid4())
    # Create an ISO8601 timestamp in the format '2019-04-29T20:54:26+00:00'
    utcnow = datetime.datetime.now(datetime.timezone.utc)
    date = utcnow.isoformat(timespec='seconds')
    for v in vals:
        v['id'] = id_
        v['date'] = date
    executemany(dbpath, sql_insert, vals)


def save_metadata(dbpath, vals):
    create_db(dbpath)
    # Bit drastic but there should only be one entry in metadata at a time
    sql_delete = "DELETE FROM metadata"
    execute(dbpath, sql_delete)

    sql_insert = "INSERT INTO metadata VALUES (:cruise, :inst)"
    executemany(dbpath, sql_insert, vals)


def save_opp_to_db(vals, dbpath):
    """
    Save aggregate statistics for filtered particle data to SQLite.

    Parameters
    ----------
    vals: list of dicts
        Values array to be saved to opp table, created by prep_opp().
    dbpath: str
        Path to SQLite DB file.
    """
    # NOTE: values inserted must be in the same order as fields in opp
    # table. Defining that order in a list here makes it easier to verify
    # that the right order is used.
    field_order = [
        "file",
        "all_count",
        "opp_count",
        "evt_count",
        "opp_evt_ratio",
        "filter_id",
        "quantile"
    ]
    values_str = ", ".join([":" + f for f in field_order])
    sql_insert = "INSERT OR REPLACE INTO opp VALUES ({})".format(values_str)
    executemany(dbpath, sql_insert, vals)


def prep_opp(file, df, all_count, evt_count, filter_id):
    """
    Prepare aggregate statistic values for filtered particle data to SQLite.

    The array returned by this function can be passed to save_opp_to_db.

    Parameters
    ----------
    file: str
        Path to SeaFlow file that was filtered. Used to get the canonical
        SeaFlow file ID.
        e.g. tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00 will become
        2014_185/2014-07-04T00-00-02+00-00.
    df: pandas.DataFrame
        SeaFlow particle data. Focused particle flag columns for each quantile
        should be in columns "q<quantile>" e.g. q2.5 for the 2.5 quantile.
    all_count: int
        Event count in raw file.
    evt_count: int
        Events above noise floor in raw file.
    filter_id: str
        DB ID for filtering parameters used to create OPP.

    Returns
    -------
    Array of values for save_opp_to_db().
    """
    vals = []
    for _q_col, q, _q_str, q_df in particleops.quantiles_in_df(df):
        opp_count = len(q_df.index)
        try:
            opp_evt_ratio = opp_count / evt_count
        except ZeroDivisionError:
            opp_evt_ratio = 0.0
        vals.append({
            "file": SeaFlowFile(file).file_id,
            "all_count": all_count,
            "opp_count": opp_count,
            "evt_count": evt_count,
            "opp_evt_ratio": opp_evt_ratio,
            "filter_id": filter_id,
            "quantile": q
        })
    return vals


def save_outlier(vals, dbpath):
    """
    Save entries in outlier table.

    Parameters
    ----------
    vals: list of dicts
        Values array to be saved to outlier table, created by prep_outlier().
    dbpath: str
        Path to SQLite DB file.
    """
    field_order = ["file", "flag"]
    values_str = ", ".join([":" + f for f in field_order])
    sql_insert = "INSERT OR REPLACE INTO outlier VALUES ({})".format(values_str)
    executemany(dbpath, sql_insert, vals)


def prep_outlier(file, flag):
    """
    Prepare an outlier entry for this file.

    Parameters
    ----------
    file: str
        Path to SeaFlow file that was filtered. Used to get the canonical
        SeaFlow file ID.
        e.g. tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00 will become
        2014_185/2014-07-04T00-00-02+00-00.
    flag: int
        Outlier table value, 0 for OK.

    Returns
    -------
    A single item array of values for save_outlier().
    """
    return [{"file": SeaFlowFile(file).file_id, "flag": flag}]


def save_sfl(dbpath, vals):
    create_db(dbpath)

    # Remove any previous SFL data
    sql_delete = "DELETE FROM sfl"
    execute(dbpath, sql_delete)

    # NOTE: values inserted must be in the same order as fields in sfl
    # table. Defining that order in a list here makes it easier to verify
    # that the right order is used.
    field_order = [
        "file",
        "date",
        "file_duration",
        "lat",
        "lon",
        "conductivity",
        "salinity",
        "ocean_tmp",
        "par",
        "bulk_red",
        "stream_pressure",
        "event_rate"
    ]
    values_str = ", ".join([":" + f for f in field_order])
    sql_insert = "INSERT OR REPLACE INTO sfl VALUES (%s)" % values_str
    executemany(dbpath, sql_insert, vals)


def get_cruise(dbpath):
    sql = "SELECT cruise FROM metadata"
    with sqlite3.connect(dbpath) as dbcon:
        df = safe_read_sql(sql, dbcon)
    if len(df.index) > 1:
        cruises = ", ".join([str(c) for c in df.cruise.tolist()])
        raise errors.SeaFlowpyError("More than one cruise found in database {}: {}".format(dbpath, cruises))
    if len(df.index) == 0:
        raise errors.SeaFlowpyError("No cruise name found in database {}\n".format(dbpath))
    return df.cruise.tolist()[0]


def get_filter_table(dbpath):
    sql = "SELECT * FROM filter ORDER BY date ASC, quantile ASC"
    with sqlite3.connect(dbpath) as dbcon:
        filterdf = safe_read_sql(sql, dbcon)
    return filterdf


def get_serial(dbpath):
    sql = "SELECT inst FROM metadata"
    with sqlite3.connect(dbpath) as dbcon:
        df = safe_read_sql(sql, dbcon)
    if len(df.index) > 1:
        insts = ", ".join([str(c) for c in df.inst.tolist()])
        raise errors.SeaFlowpyError("More than one instrument serial found in database {}: {}".format(dbpath, insts))
    if len(df.index) == 0:
        raise errors.SeaFlowpyError("No instrument serial found in database {}\n".format(dbpath))
    return df.inst.tolist()[0]


def get_latest_filter(dbpath):
    with sqlite3.connect(dbpath) as dbcon:
        df = safe_read_sql("SELECT * FROM filter ORDER BY date DESC, quantile ASC", dbcon)
    if len(df.index) == 0:
        raise errors.SeaFlowpyError("No filter parameters found in database {}\n".format(dbpath))
    _id = df.iloc[0]["id"]
    return df[df["id"] == _id]


def get_filter_params_lookup(dbpath, files_df):
    files_df = files_df.copy().reset_index(drop=False)
    with sqlite3.connect(dbpath) as dbcon:
        filter_df = safe_read_sql("SELECT * FROM filter ORDER BY date DESC, quantile ASC", dbcon)
        filter_plan_df = safe_read_sql("SELECT * FROM filter_plan ORDER BY start_date ASC", dbcon)
    if len(filter_df.index) == 0:
        raise errors.SeaFlowpyError("No filter parameters found in database {}\n".format(dbpath))
    if len(filter_plan_df.index) == 0:
        raise errors.SeaFlowpyError("No filter plan found in database {}\n".format(dbpath))

    files_df["filter_id"] = None
    if len(filter_plan_df) > 1:
        for i in range(len(filter_plan_df) - 1):
            gte = files_df["date"] >= filter_plan_df.loc[i, "start_date"]
            lte = files_df["date"] <= filter_plan_df.loc[i + 1, "start_date"]
            files_df.loc[(gte & lte), "filter_id"] = filter_plan_df.loc[i, "filter_id"]
    i = len(filter_plan_df) - 1
    gte = files_df["date"] >= filter_plan_df.loc[i, "start_date"]
    files_df.loc[gte, "filter_id"] = filter_plan_df.loc[i, "filter_id"]

    filter_params = {}
    for i, row in files_df.iterrows():
        filter_params[row["file_id"]] = filter_df[filter_df["id"] == row["filter_id"]].reset_index(drop=True)

    return filter_params


def get_opp_table(dbpath, filter_id=""):
    if filter_id == "":
        sql = "SELECT * FROM opp ORDER BY file ASC, quantile ASC"
    else:
        sql = "SELECT * FROM opp WHERE filter_id = '{}' ORDER BY file ASC, quantile ASC".format(filter_id)
    with sqlite3.connect(dbpath) as dbcon:
        oppdf = safe_read_sql(sql, dbcon)
    return oppdf


def get_outlier_table(dbpath):
    sql = "SELECT * FROM outlier ORDER BY file ASC"
    with sqlite3.connect(dbpath) as dbcon:
        outlierdf = safe_read_sql(sql, dbcon)
    return outlierdf


def get_sfl_table(dbpath):
    sql = "SELECT * FROM sfl ORDER BY date ASC"
    with sqlite3.connect(dbpath) as dbcon:
        df = safe_read_sql(sql, dbcon)
    return df


def get_event_counts(dbpath):
    filterid = get_latest_filter(dbpath).loc[0, "id"]
    opp = get_opp_table(dbpath, filterid)
    grouped = opp[["file", "all_count"]].groupby(["file"])
    return {name: group["all_count"].head(1).values[0] for name, group in grouped}


def merge_dbs(db1, db2):
    """Merge two SQLite databases into a new database."""
    with sqlite3.connect(db1) as con1:
        with sqlite3.connect(db2) as con2:
            gatingdf = safe_read_sql('select * from gating', con1)
            polydf = safe_read_sql('select * from poly', con1)
            filterdf = safe_read_sql('select * from filter', con1)
            gatingdf.to_sql('gating', con2, if_exists='append', index=False)
            polydf.to_sql('poly', con2, if_exists='append', index=False)
            filterdf.to_sql('filter', con2, if_exists='append', index=False)
    # Merge opp
    # Merge vct
    # Merge meta


def execute(dbpath, sql, timeout=120):
    con = sqlite3.connect(dbpath, timeout=timeout)
    try:
        with con:
            con.execute(sql)
    except sqlite3.Error as e:
        raise errors.SeaFlowpyError("An error occurred when executing SQL queries: {!s}".format(e))
    finally:
        con.close()


def executemany(dbpath, sql, values=None, timeout=120):
    con = sqlite3.connect(dbpath, timeout=timeout)
    try:
        with con:
            con.executemany(sql, values)
    except sqlite3.Error as e:
        raise errors.SeaFlowpyError("An error occurred when executing SQL queries: {!s}".format(e))
    finally:
        con.close()

def executescript(dbpath, sql_script_text, timeout=120):
    con = sqlite3.connect(dbpath, timeout=timeout)
    try:
        with con:
            con.executescript(sql_script_text)
    except sqlite3.Error as e:
        raise errors.SeaFlowpyError("An error occurred when executing a SQL script: {!s}".format(e))
    finally:
        con.close()


def safe_read_sql(sql, con):
    """Catch and handle error if table not present during pandas.read_sql()"""
    try:
        df = pd.read_sql(sql, con)
        errmsg = ''
    except pd.io.sql.DatabaseError as e:
        errmsg = str(e)
    if errmsg:
        raise errors.SeaFlowpyError(errmsg)
    return df



import datetime
from builtins import str
import click
from seaflowpy import errors


FILE_COLUMNS = [
    'FILE', 'DATE', 'FILE DURATION', 'LAT', 'LON', 'CONDUCTIVITY',
    'SALINITY', 'OCEAN TEMP', 'PAR', 'BULK RED', 'STREAM PRESSURE',
    'EVENT RATE'
]

def parse_header(line):
    """Create a dict of column indexes by field name"""
    fields = line.rstrip().split('\t')
    fields = [f.replace('.', ' ') for f in fields]
    d = {x:i for i, x in enumerate(fields)}
    return d

def create_file_field(line, header):
    """Create SeaFlow file name form this line"""
    file_field = None
    fields = line.rstrip().split('\t')
    if ('file' in header) and ('day' in header):
        filei = header['file']
        dayi = header['day']
        file_field = fields[dayi] + '/' + fields[filei] + '.evt'
    elif 'FILE' in header:
        filei = header['FILE']
        if fields[filei].startswith('sds_'):
            parts = fields[filei].split('_')
            file_field = parts[1] + '_' + parts[2] + '/' + parts[3] + '.evt'
    if not file_field:
        raise errors.FileError('Could not create file name from line:\n%s' % line)
    return file_field

def create_date_field(line, header):
    """Create ISO8601 date value from this line"""
    date_field = None
    fields = line.rstrip().split('\t')
    if 'time' in header:
        timei = header['time']
        date_field = fields[timei].replace(' ', 'T') + '+00:00'  # assume UTC
    elif 'computerUTC' in header:
        # Given the following definitions:
        # y = year
        # j = day of year
        # h = hour
        # m = minute
        # s = second
        #
        # Then computerUTC is formatted as yyjjjhhmmss
        cutc = fields[header['computerUTC']]
        assert len(cutc) == len('yyjjjhhmmss')
        year = int('20' + cutc[:2])
        dayofyear = int(cutc[2:5])
        hours = int(cutc[5:7])
        minutes = int(cutc[7:9])
        seconds = int(cutc[9:11])

        # dayofyear to month/day
        delta = datetime.timedelta(days=dayofyear-1, hours=hours, minutes=minutes,
                                   seconds=seconds)
        tmpd = datetime.datetime(year, 1, 1) + delta
        # Final UTC ISO8601 SeaFlow compatible timestamp string
        date_field = tmpd.isoformat()  + '+00:00'
    if not date_field:
        raise errors.FileError('could not create date from line:\n%s' % line)
    return date_field

@click.command()
@click.argument('input-sds', type=click.File())
@click.argument('output-sfl', type=click.File(mode='w', atomic=True))
def sds2sfl_cmd(input_sds, output_sfl):
    """Convert SDS file format to SFL."""
    header = parse_header(input_sds.readline())

    output_sfl.write('\t'.join(FILE_COLUMNS) + '\n')

    missing_fields = {}

    file_duration_field = '180'
    for line in input_sds:
        try:
            file_field = create_file_field(line, header)
            date_field = create_date_field(line, header)
        except errors.FileError as e:
            raise click.ClickException(str(e))

        # Make a copy of header index lookup
        d = dict(header)
        # Capture original fields
        fields = line.rstrip().split('\t')
        # Add new or modified fields
        fields.append(file_duration_field)
        d['FILE DURATION'] = len(fields) - 1
        fields.append(file_field)
        d['FILE'] = len(fields) - 1
        fields.append(date_field)
        d['DATE'] = len(fields) - 1


        # Write SFL subset of fields in correct order
        outfields = []
        for col in FILE_COLUMNS:
            try:
                outfields.append(fields[d[col]])
            except KeyError as e:
                # If column is missing, just output NA and note for later
                outfields.append('NA')
                missing_fields[str(e)] = True

        output_sfl.write('\t'.join(outfields) + '\n')

    if missing_fields:
        click.echo('Some fields were missing from input file: %s' % ' '.join(list(missing_fields.keys())))



import datetime
import os

import numpy as np
import pandas as pd
import pytest
import pytz
import seaflowpy as sfp

# pylint: disable=redefined-outer-name

@pytest.fixture()
def tmpout(tmpdir):
    """Setup to test sampling workflow"""
    evtpaths = [
        "tests/testcruise_evt/2014_185/2014-07-04T00-00-02+00-00",
        "tests/testcruise_evt/2014_185/2014-07-04T00-03-02+00-00.gz",
    ]
    file_ids = [
        "2014_185/2014-07-04T00-00-02+00-00",
        "2014_185/2014-07-04T00-03-02+00-00",
    ]
    dates = [
        datetime.datetime.fromisoformat("2014-07-04T00:00:02+00:00"),
        datetime.datetime.fromisoformat("2014-07-04T00:03:02+00:00"),
    ]
    dates = [d.replace(tzinfo=pytz.utc) for d in dates]
    dates_lookup = dict(zip(file_ids, dates))

    return {
        "evtpaths": evtpaths,
        "file_ids": file_ids,
        "tmpdir": str(tmpdir),
        "dates": dates_lookup
    }


class TestSample:
    def test_sample_evt_single(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "test.gz")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"], 20000, outpath, noise_filter=False, seed=12345
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        assert [r["events"] for r in results] == [40000, 40000]
        assert [r["events_postfilter"] for r in results] == [40000, 40000]
        assert [r["events_postsampling"] for r in results] == [10000, 10000]
        df = pd.read_parquet(outpath)
        assert len(df.index) == 20000
        assert len(df[(df["D1"] == 0) & (df["D2"] == 0) & (df["fsc_small"] == 0)]) > 0

    def test_sample_evt_single_noise_filter(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "test.gz")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"], 20000, outpath, noise_filter=True, seed=12345
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        assert [r["events"] for r in results] == [40000, 40000]
        assert [r["events_postfilter"] for r in results] == [39928, 39925]
        assert [r["events_postsampling"] for r in results] == [10000, 10000]
        df = pd.read_parquet(outpath)
        assert len(df.index) == 20000
        assert len(df[(df["D1"] == 0) & (df["D2"] == 0) & (df["fsc_small"] == 0)]) == 0

    def test_sample_evt_single_min_filter(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "test.gz")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"],
            20000,
            outpath,
            min_fsc=1,
            min_chl=25000,
            min_pe=25000,
            seed=12345,
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        assert [r["events"] for r in results] == [40000, 40000]
        assert [r["events_postfilter"] for r in results] == [379, 471]
        assert [r["events_postsampling"] for r in results] == [379, 471]
        df = pd.read_parquet(outpath)
        assert len(df.index) == 850
        assert np.min(df["fsc_small"]) >= 1
        assert np.min(df["pe"]) >= 25000
        assert np.min(df["chl_small"]) >= 25000

    def test_sample_evt_single_empty(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "test.gz")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"],
            20000,
            outpath,
            min_fsc=60000,
            min_chl=60000,
            min_pe=60000,
            seed=12345,
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        assert [r["events"] for r in results] == [40000, 40000]
        assert [r["events_postfilter"] for r in results] == [0, 0]
        assert [r["events_postsampling"] for r in results] == [0, 0]
        df = pd.read_parquet(outpath)
        assert len(df.index) == 0

    def test_sample_evt_single_dates(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "test.gz")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"],
            20000,
            outpath,
            dates=tmpout["dates"],
            noise_filter=False,
            seed=12345
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        df = pd.read_parquet(outpath)
        assert len(df) == 20000
        assert len(df["date"].unique()) == 2
        assert len(df.head(10000)["date"].unique()) == 1
        assert df.head(10000)["date"].unique()[0].isoformat() == "2014-07-04T00:00:02+00:00"
        assert df.tail(10000)["date"].unique()[0].isoformat() == "2014-07-04T00:03:02+00:00"

    def test_sample_evt_multi(self, tmpout):
        outpath = os.path.join(tmpout["tmpdir"], "testdir")
        results, errs = sfp.sample.sample(
            tmpout["evtpaths"], 20000, outpath, multi=True, seed=12345
        )
        assert len(errs) == 0
        assert len(results) == 2
        assert [r["file_id"] for r in results] == tmpout["file_ids"]
        assert [r["events"] for r in results] == [40000, 40000]
        assert [r["events_postfilter"] for r in results] == [40000, 40000]
        assert [r["events_postsampling"] for r in results] == [20000, 20000]

        df = pd.read_parquet(outpath)
        assert (
            len(df[(df["D1"] == 0) & (df["D2"] == 0) & (df["fsc_small"] == 0)]) > 0
        )
        gb = df.groupby("file_id")
        assert gb.ngroups == 2
        assert list(gb.groups.keys()) == tmpout["file_ids"]
        assert [len(g) for g in gb.groups.values()] == [20000, 20000]



import json
import click
import pandas as pd
import pkg_resources
from seaflowpy import db
from seaflowpy import errors
from seaflowpy import filterevt
from seaflowpy import util
from seaflowpy import seaflowfile


def validate_limit(ctx, param, value):
    if value is not None and value < 1:
        raise click.BadParameter('if limit is set, it must be >= 1')
    return value


def validate_process_count(ctx, param, value):
    if value < 1:
        raise click.BadParameter('process_count must be >= 1')
    return value


def validate_resolution(ctx, param, value):
    if value <= 0 or value > 100:
        raise click.BadParameter('resolution must be a number between 1 and 100 inclusive.')
    return value


@click.command()
@click.option('-D', '--delta', is_flag=True,
    help='Filter EVT files which are not already present in the opp table.')
@click.option('-e', '--evt-dir', metavar='DIR', type=click.Path(exists=True), required=True,
    help='EVT directory path (required unless --s3)')
@click.option('-d', '--db', 'dbpath', required=True, metavar='FILE', type=click.Path(exists=True),
    help='Popcycle SQLite3 db file with filter parameters and cruise name.')
@click.option('-l', '--limit', type=int, metavar='N', callback=validate_limit,
    help='Limit number of files to process.')
@click.option('-o', '--opp-dir', metavar='DIR',
    help='Directory in which to save OPP files. Will be created if does not exist.')
@click.option('-p', '--process-count', default=1, show_default=True, metavar="N", callback=validate_process_count,
    help='Number of processes to use in filtering.')
@click.option('-r', '--resolution', default=10.0, show_default=True, metavar='N', callback=validate_resolution,
    help='Progress update resolution by %%.')
@util.quiet_keyboardinterrupt
def filter_cmd(delta, evt_dir, dbpath, limit, opp_dir, process_count, resolution):
    """Filter EVT data locally."""
    # Find cruise in db
    try:
        cruise = db.get_cruise(dbpath)
    except errors.SeaFlowpyError as e:
        raise click.ClickException(str(e))

    # Find filter parameters in db. Won't use them yet but better to check
    # upfront
    try:
        _filter_params = db.get_latest_filter(dbpath)
    except errors.SeaFlowpyError as e:
        raise click.ClickException(str(e))

    # Capture run parameters and information
    v = {
        'delta': delta,
        'evt_dir': evt_dir,
        'limit': limit,
        'db': dbpath,
        'opp_dir': opp_dir,
        'process_count': process_count,
        'resolution': resolution,
        'version': pkg_resources.get_distribution("seaflowpy").version,
        'cruise': cruise
    }
    to_delete = [k for k in v if v[k] is None]
    for k in to_delete:
        v.pop(k, None)  # Remove undefined parameters

    # Print run parameters
    print('Run parameters and information:')
    print(json.dumps(v, indent=2))
    print('')

    # Find EVT files
    print('Getting lists of files to filter')
    evt_files = seaflowfile.sorted_files(seaflowfile.find_evt_files(evt_dir))

    # Check for duplicates, exit with message if any exist
    # This could be caused by gzipped and uncompressed files in the same location
    uniques = {seaflowfile.SeaFlowFile(f).file_id for f in evt_files}
    if len(uniques) < len(evt_files):
        raise click.ClickException('Duplicate EVT file(s) detected')

    # Get DataFrame of file IDs, paths, dates in common between discovered
    # EVT file paths and SFL files.
    try:
        sfl_df = db.get_sfl_table(dbpath)
    except (errors.SeaFlowpyError, KeyError, ValueError) as e:
        raise click.ClickException(str(e))
    files_df = seaflowfile.date_evt_files(evt_files, sfl_df)

    # Find intersection of SFL files and EVT files
    print('sfl={} evt={} intersection={}'.format(len(sfl_df), len(evt_files), len(files_df)))

    if delta:
        # Limit files to those not present in the opp table
        try:
            opp_df = db.get_opp_table(dbpath)
            opp_df = opp_df.rename(columns={"file": "file_id"}).loc[:, "file_id"]
        except (errors.SeaFlowpyError, KeyError, ValueError) as e:
            raise click.ClickException(str(e))
        # Find files in common
        merged = pd.merge(files_df, opp_df, how='left', on=["file_id"], indicator=True)
        evt_n = len(files_df)
        files_df = merged[merged["_merge"] == "left_only"]
        print('opp={} evt={} evt_not_in_opp={}'.format(len(opp_df), evt_n, len(files_df)))

    # Restrict length of file list with --limit
    if (limit is not None) and (limit > 0):
        files_df = files_df.head(limit)

    if len(files_df.index) > 0:
        # Filter
        try:
            filterevt.filter_evt_files(
                files_df,
                dbpath,
                opp_dir,
                worker_count=process_count,
                every=resolution
            )
        except errors.SeaFlowpyError as e:
            raise click.ClickException(str(e))



import copy
#import datetime
#import os
import sys
import time
import multiprocessing as mp
import queue

from . import db
from . import errors
from . import fileio
from . import particleops
from . import util


# Stop sentinel value for queues
stop = 'STOP'
# Quantile list
quantiles = [2.5, 50, 97.5]


@util.quiet_keyboardinterrupt
def filter_evt_files(files_df, dbpath, opp_dir, worker_count=1, every=10.0,
                     window_size="1H"):
    """Filter a list of EVT files.

    Positional arguments:
        files_df - DataFrame of "file_id", "path", "date" as file ID string,
            file path, and pandas.Timestamp for the file.
        dbpath = SQLite3 db path
        opp_dir = Directory for output binary OPP files

    Keyword arguments:
        worker_count - number of worker processes to use
        every - Percent progress output resolution
        window_size - Time window for grouping filtering EVT file sets,
            expressed as pandas time offsets.
    """
    work = {
        "files_df": None,  # fill in later
        "cloud_config_items": None,
        "dbpath": dbpath,
        "opp_dir": opp_dir,
        "filter_params": None,  # fill in later from db,
        "window_size": window_size,
        "window_start_date": None,
        "errors": [],  # global errors outside of processing single files
        "results": []
    }

    if not dbpath:
        raise ValueError("Must provide db path to filter_evt_files()")
    if worker_count < 1:
        raise ValueError("worker_count must be > 0")
    if every <= 0 or every > 100:
        raise ValueError("resolution must be > 0 and <= 100")

    # Group by window_size
    grouped = files_df.set_index("date").resample(window_size)

    worker_count = min(len(grouped), worker_count)

    # Get a dictionary of file_id: filter_params
    work["filter_params"] = db.get_filter_params_lookup(dbpath, files_df)

    # Create input queue with info necessary to filter one file
    work_q = mp.Queue()
    # Create output queues
    stats_q = mp.Queue()  # result stats
    opps_q = mp.Queue()   # OPP data
    done_q = mp.Queue()   # signal we're done to main thread

    # Create worker processes
    workers = []
    for _ in range(worker_count):
        p = mp.Process(target=do_filter, args=(work_q, opps_q))
        p.start()
        workers.append(p)

    # Create db output process
    saver = mp.Process(
        target=do_save,
        args=(opps_q, stats_q, len(files_df))
    )
    saver.start()

    # Create reporting process
    reporter = mp.Process(
        target=do_reporting,
        args=(stats_q, done_q, len(files_df), every)
    )
    reporter.start()

    # Add work to the work queue, binned by hour
    for name, group in grouped:
        if len(group) > 0:
            work_copy = copy.deepcopy(work)
            work_copy["files_df"] = group.copy()
            work_copy["window_start_date"] = name
            work_q.put(work_copy)
    # Put sentinel stop values on the input queue, one for each consumer process
    for _ in range(worker_count):
        work_q.put(stop)

    try:
        # Wait for reporter to tell us we're done
        done = done_q.get()
        if done is not None:
            # Something went wrong, shut child processes down
            print(done, file=sys.stderr)
    finally:
        for w in workers:
            w.terminate()
            w.join()
        saver.terminate()
        saver.join()
        reporter.join()


@util.quiet_keyboardinterrupt
def do_filter(work_q, opps_q):
    """Filter one EVT file, save to sqlite3, return filter stats"""
    work = work_q.get()
    while work != stop:
        #print("{} {} starting {} at {}".format(work["window_start_date"], os.getpid(), len(work["files_df"]), datetime.datetime.now().isoformat()), file=sys.stderr)
        for date, row in work["files_df"].iterrows():
            result = {
                "error": "",
                "all_count": 0,
                "evt_count": 0,
                "saturated_count": 0,
                "opp": None,
                "file_id": row["file_id"],
                "path": row["path"]
            }

            filter_params = work["filter_params"][row["file_id"]].reset_index(drop=True)
            try:
                data = fileio.read_evt_labview(path=row["path"])
                evt_df = data["df"]
            except errors.FileError as e:
                result["error"] = f"Could not parse file {row['path']}: {e}"
                evt_df = particleops.empty_df()  # doesn't matter if v1 or v2 column composition
            except Exception as e:
                result["error"] = f"Unexpected error when parsing file {row['path']}: {e}"
                evt_df = particleops.empty_df()  # doesn't matter if v1 or v2 column composition

            try:
                evt_df = particleops.mark_focused(evt_df, filter_params, inplace=True)
                opp_df = particleops.select_focused(evt_df)
            except Exception as e:
                result["error"] = f"Unexpected error when marking and selecting focused particles in file {row['path']}: {e}"
            else:
                opp_df["date"] = date
                opp_df["file_id"] = row["file_id"]
                opp_df["filter_id"] = filter_params["id"][0]
                result["opp"] = opp_df
                result["all_count"] = len(evt_df.index)
                result["noise_count"] = len(evt_df[evt_df["noise"]].index)
                result["saturated_count"] = len(evt_df[evt_df["saturated"]].index)
                result["opp_count"] = len(opp_df[opp_df["q50"]])
                result["filter_id"] = filter_params["id"][0]

            work["results"].append(result)

        # Prep db data
        work["opp_vals"], work["outlier_vals"] = [], []
        for r in work["results"]:
            work["opp_vals"].extend(
                db.prep_opp(
                    r["file_id"],
                    r["opp"],
                    r["all_count"],
                    r["all_count"] - r["noise_count"],
                    r["filter_id"]
                )
            )

            work["outlier_vals"].extend(db.prep_outlier(r["file_id"], 0))
        # Save OPP file
        # Only include OPP files with data in all quantiles
        good_opps = []
        for r in work["results"]:
            if (not r["error"]) and particleops.all_quantiles(r["opp"]):
                good_opps.append(r["opp"])
        if (len(good_opps)):
            #print("{} {} saving parquet at {}".format(work["window_start_date"], os.getpid(), datetime.datetime.now().isoformat()), file=sys.stderr)
            try:
                if work["opp_dir"]:
                    fileio.write_opp_parquet(
                        good_opps,
                        work["window_start_date"],
                        work["window_size"],
                        work["opp_dir"]
                    )
            except Exception as e:
                work["errors"].append(f"Unexpected error when saving OPP for {work['window_start_date']}: {e}")
        else:
            work["errors"].append(f"No OPPs had data in all quantiles for {work['window_start_date']}")

        # Erase OPP from payload
        for r in work["results"]:
            del r["opp"]

        #print("{} {} sending {}/{} results at {}".format(work["window_start_date"], os.getpid(), len(work["results"]), len(work["files_df"]), datetime.datetime.now().isoformat()), file=sys.stderr)
        opps_q.put(work)
        #print("{} {} sent {}/{} results at {}".format(work["window_start_date"], os.getpid(), len(work["results"]), len(work["files_df"]), datetime.datetime.now().isoformat()), file=sys.stderr)
        work = work_q.get()
        # if work != stop:
        #     print("{} {} got more work {} at {}".format(work["window_start_date"], os.getpid(), len(work["files_df"]), datetime.datetime.now().isoformat()), file=sys.stderr)
        # else:
        #     print("{} got stop {}".format(os.getpid(), datetime.datetime.now().isoformat()), file=sys.stderr)


@util.quiet_keyboardinterrupt
def do_save(opps_q, stats_q, files_left):
    while files_left > 0:
        try:
            work = opps_q.get(True, 600)  # We should get one hour of data every ten minutes at least
            #print("{} {} received {}/{} results at {}".format(work["window_start_date"], os.getpid(), len(work["results"]), len(work["files_df"]), datetime.datetime.now().isoformat()), file=sys.stderr)
        except queue.Empty:
            stats_q.put("EMPTY QUEUE")
            break
        except Exception:
            stats_q.put("QUEUE ERROR")
            break

        files_left -= len(work["files_df"])

        # Save to DB
        try:
            if work["dbpath"]:
                if work["opp_vals"]:
                    db.save_opp_to_db(work["opp_vals"], work["dbpath"])
                if work["outlier_vals"]:
                    db.save_outlier(work["outlier_vals"], work["dbpath"])
                #print("{} {} db saved at {}".format(work["window_start_date"], os.getpid(), datetime.datetime.now().isoformat()), file=sys.stderr)
        except Exception as e:
            work["errors"].append("Unexpected error when saving to db for {} batch starting at {}: {}".format(work["window_size"], work["window_start_date"], e))

        #print("{} {} sent stats at {}".format(work["window_start_date"], os.getpid(), datetime.datetime.now().isoformat()), file=sys.stderr)
        stats_q.put(work)


@util.quiet_keyboardinterrupt
def do_reporting(stats_q, done_q, file_count, every):
    event_count = 0
    noise_count = 0
    signal_count = 0
    saturated_count = 0
    opp_count = 0
    files_ok = 0

    print("")
    print(f"Filtering {file_count} EVT files. Progress for 50th quantile every ~ {every}%")

    t0 = time.time()

    last = 0  # Last progress milestone in increments of every
    event_count_block = 0  # EVT particles in this block (between milestones)
    noise_count_block = 0  # EVT noise particles in this block
    signal_count_block = 0  # EVT signal (not noise) particles in this block
    saturated_count_block = 0  # particles saturating D1 or D2
    opp_count_block = 0  # OPP particles in this block
    files_seen = 0
    files_left = file_count

    # Filter particles in parallel with process pool
    while files_left > 0:
        work = stats_q.get()  # get next result group

        if work in ("EMPTY QUEUE", "QUEUE ERROR"):
            # Something went wrong upstream, exit with an error message
            done_q.put(f"A fatal error occurred after filtering {files_seen}/{file_count} files: {work}")
            sys.exit(1)

        #print("{} {} received stats at {}".format(work["window_start_date"], os.getpid(), time.time()), file=sys.stderr)

        files_left -= len(work["files_df"])

        if work["errors"]:
            for e in work["errors"]:
                print(e, file=sys.stderr)

        for r in work["results"]:
            files_seen += 1

            if r["error"]:
                print(r["error"], file=sys.stderr)
            else:
                files_ok += 1

            event_count_block += r["all_count"]
            noise_count_block += r["noise_count"]
            signal_count_block = event_count_block - noise_count_block
            saturated_count_block += r["saturated_count"]
            opp_count_block += r["opp_count"]

            # Print progress periodically
            perc = float(files_seen) / file_count * 100  # Percent completed
            # Round down to closest every%
            milestone = int(perc / every) * every
            if milestone > last:
                event_count += event_count_block
                noise_count += noise_count_block
                signal_count += signal_count_block
                saturated_count += saturated_count_block
                opp_count += opp_count_block
                ratio_noise_block = util.zerodiv(noise_count_block, event_count_block)
                ratio_saturated_block = util.zerodiv(saturated_count_block, event_count_block)
                ratio_evtopp_block = util.zerodiv(opp_count_block, signal_count_block)
                msg = f"File: {files_seen}/{file_count} {perc:5.4}%"
                msg += " events: %d noise: %d (%.04f) sat: %d (%.04f) opp: %d (%.04f) t: %.2fs" % \
                    (
                        event_count_block,
                        noise_count_block, ratio_noise_block,
                        saturated_count_block, ratio_saturated_block,
                        opp_count_block, ratio_evtopp_block,
                        time.time() - t0
                    )
                print(msg)
                sys.stdout.flush()
                last = milestone
                event_count_block = 0
                noise_count_block = 0
                signal_count_block = 0
                saturated_count_block = 0
                opp_count_block = 0

    # If any particle count data is left, add it to totals
    event_count += event_count_block
    noise_count += noise_count_block
    signal_count += signal_count_block
    saturated_count += saturated_count_block
    opp_count += opp_count_block

    ratio_noise = util.zerodiv(noise_count, event_count)
    ratio_saturated = util.zerodiv(saturated_count, event_count)
    ratio_evtopp = util.zerodiv(opp_count, signal_count)

    summary_text = "Total events: %d noise: %d (%.04f) sat: %d (%.04f) opp: %d (%.04f) t: %.2fs" % \
        (
            event_count,
            noise_count, ratio_noise,
            saturated_count, ratio_saturated,
            opp_count, ratio_evtopp,
            time.time() - t0
        )
    print(summary_text)
    print(f"{files_ok} / {file_count} EVT files parsed successfully")
    done_q.put(None)



"""Module for common SeaFlow datetime operations."""
import datetime
import pytz


def parse_date(date_str, assume_utc=True):
    """
    Parse a SeaFlow timestamp.

    If assume_utc, ignore any timezone and assume UTC. If no timezone offset is
    found in date_str, always set to UTC.

    Returns a datetime.datetime object. Raises ValueError if the timestamp can't
    be parsed.
    """
    # First normalize timezone Z to +00:00
    if date_str.endswith("Z"):
        date_str = date_str[:-1] + "+00:00"
    date = datetime.datetime.fromisoformat(date_str)
    if assume_utc or date.tzinfo is None:
        date = date.replace(tzinfo=pytz.utc)
    return date


def seaflow_rfc3339(date):
    """Standard SeaFlow RFC3339 timestamp."""
    return date.isoformat(timespec='seconds')



from functools import wraps
from signal import getsignal, signal, SIGPIPE, SIG_DFL
import errno
import os
import subprocess
import sys
import time


def find_files(root_dir):
    """Return a list of all file paths below root_dir."""
    if not os.path.isdir(root_dir):
        raise OSError("directory does not exist")
    allfiles = []
    def error_handler(e):
        raise e
    for root, _dirs, files in os.walk(root_dir, onerror=error_handler):
        for f in files:
            allfiles.append(os.path.join(root, f))
    return allfiles


def gzip_file(path, print_timing=False):
    """Gzip a file.

    Try to use pigz, but fall back to gzip.
    """
    gzipbin = "pigz"  # Default to using pigz
    devnull = open(os.devnull, "w")
    try:
        subprocess.check_call(["pigz", "--version"], stdout=devnull,
                              stderr=subprocess.STDOUT)
    except OSError:
        # If pigz is not installed fall back to gzip
        gzipbin = "gzip"

    if print_timing:
        t0 = time.time()
        print("")
        print("Compressing %s" % path)

    subprocess.check_call([gzipbin, "-f", path],
                          stderr=subprocess.STDOUT)

    if print_timing:
        t1 = time.time()
        print("Compression completed in %.2f seconds" % (t1 - t0))


def jobs_parts(things, n):
    """Split a list of things into n sublists."""
    if n < 1:
        raise ValueError("n must be > 1")
    n = int(min(len(things), n))
    per_part = len(things) // n
    rem = len(things) % n
    buckets = []
    i = 0
    while i < len(things):
        if rem > 0:
            extra = 1
        else:
            extra = 0
        start = i
        end = i + per_part + extra
        rem -= 1
        i = end
        buckets.append(things[start:end])
    return buckets


def mkdir_p(path):
    """Create directory tree for path."""
    if path == '':
        return
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def quantile_str(q):
    """
    Display quantile float as string.

    If there is not decimal part, don't display ".0". If there is a decimal
    part, display it.
    """
    return "{0}".format(str(q) if q % 1 else int(q))


def splitpath(path):
    """Return a list of all path components"""
    parts = []
    path, last = os.path.split(path)
    if last != "":
        parts.append(last)
    while True:
        path, last = os.path.split(path)
        if last != "":
            parts.append(last)
        else:
            if path != "":
                parts.append(path)
            break
    return parts[::-1]


def suppress_sigpipe(f):
    """Decorator to handle SIGPIPE cleanly.

    Prevent Python from turning SIGPIPE into an exception and printing an
    uncatchable error message. Note, if the wrapped function depends on the
    default behavior of Python when handling SIGPIPE this decorator may have
    unintended effects."""
    @wraps(f)
    def wrapper(*args, **kwargs):
        orig_handler = getsignal(SIGPIPE)
        signal(SIGPIPE, SIG_DFL)
        try:
            f(*args, **kwargs)
        finally:
            signal(SIGPIPE, orig_handler)  # restore original Python SIGPIPE handler
    return wrapper


def quiet_keyboardinterrupt(f):
    """Decorator to exit quietly on keyboard interrupt."""
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except KeyboardInterrupt:
            sys.exit()
    return wrapper


def zerodiv(x, y):
    """Divide x by y, floating point, and default to 0.0 if divisor is 0"""
    try:
        answer = float(x) / float(y)
    except ZeroDivisionError:
        answer = 0.0
    return answer



import numpy as np
import pandas as pd
from . import util


# Data columns in raw SeaFlow particle DataFrame
COLUMNS = [
    "time", "pulse_width", "D1", "D2", "fsc_small", "fsc_perp", "fsc_big",
    "pe", "chl_small", "chl_big"
]
CHANNEL_COLUMNS = COLUMNS[2:]  # flow cytometer channel data columns
COLUMNS2 = [
    "pulse_width", "chl_small", "D1", "D2", "fsc_small", "pe", "evt_rate"
]
CHANNEL_COLUMNS2 = COLUMNS[1:6]  # flow cytometer channel data columns

# Focused particle masks by quantile column. These get combined into bit flags
# when storing OPP data in a binary file. e.g. 0b110 (6) means a particle is
# focused in quantiles 50.0 and 97.5 but not 2.5.
flags = {
    "q2.5": 1,
    "q50": 2,
    "q97.5": 4
}


def all_quantiles(df):
    """
    Are there particles in all quantiles?

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow particle data with focused particles marked by mark_focused().
        Focused particles should be marked with a boolean column for each
        quantile, where column names are q<quantile>, e.g. q2.5 for 2.5%
        quantile.

    Returns
    ------
    bool
    """
    for q_col in [c for c in df.columns if c.startswith("q")]:
        if not df[q_col].any():
            return False
    return True


def empty_df(v2=False):
    """
    Create an empty SeaFlow particle DataFrame.

    Returns
    -------
    pandas.DataFrame
    """
    cols = COLUMNS2 if v2 else COLUMNS
    return pd.DataFrame(dtype=float, columns=cols)


def mark_focused(df, params, inplace=False):
    """
    Mark focused particle data.

    Adds boolean cols for noise, saturation, and focused particles by quantile.

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow raw event DataFrame.
    params: pandas.DataFrame
        Filtering parameters as pandas DataFrame.
    inplace: bool, default False
        Add new booleans columns to and return input DataFrame. If False,
        add new columns to and return a copy of the input DataFrame, leaving the
        original unmodified.

    Returns
    -------
    pandas.DataFrame
        Reference to or copy of input DataFrame with new boolean columns.
    """
    # Check parameters
    param_keys = [
        "width", "notch_small_D1", "notch_small_D2", "notch_large_D1",
        "notch_large_D2", "offset_small_D1", "offset_small_D2",
        "offset_large_D1", "offset_large_D2", "quantile"
    ]
    if params is None:
        raise ValueError("Must provide filtering parameters")
    for k in param_keys:
        if not k in params.columns:
            raise ValueError(f"Missing filter parameter {k} in mark_focused")
    # Make sure params have 0-based indexing
    params = params.reset_index(drop=True)

    if not inplace:
        df = df.copy()

    # Apply noise filter and saturation filter for D1 and D2
    df["noise"] = mark_noise(df)
    df["saturated"] = mark_saturated(df)

    # Filter for aligned/focused particles
    #
    # Filter aligned particles (D1 = D2), with correction for D1 D2
    # sensitivity difference. Assume width is same for all quantiles so just
    # grab first width value and calculate aligned particles once
    assert len(params["width"].unique()) == 1  # may as well check
    width = params.loc[0, "width"]
    alignedD1 = df["D1"].values < (df["D2"].values + width)
    alignedD2 = df["D2"].values < (df["D1"].values + width)
    aligned = ~df["noise"] & ~df["saturated"] & alignedD1 & alignedD2

    for q in params["quantile"].sort_values():
        p = params[params["quantile"] == q].iloc[0]  # get first row of dataframe as series
        # Filter focused particles
        # Using underlying numpy arrays (values) to construct boolean
        # selector is about 10% faster than using pandas Series
        small_D1 = df["D1"].values <= ((df["fsc_small"].values * p["notch_small_D1"]) + p["offset_small_D1"])
        small_D2 = df["D2"].values <= ((df["fsc_small"].values * p["notch_small_D2"]) + p["offset_small_D2"])
        large_D1 = df["D1"].values <= ((df["fsc_small"].values * p["notch_large_D1"]) + p["offset_large_D1"])
        large_D2 = df["D2"].values <= ((df["fsc_small"].values * p["notch_large_D2"]) + p["offset_large_D2"])
        opp_selector = aligned & ((small_D1 & small_D2) | (large_D1 & large_D2))
        # Mark focused particles
        colname = f"q{util.quantile_str(q)}"
        df[colname] = opp_selector

    return df


def mark_noise(df):
    """
    Mark data below noise threshold.

    This function returns a boolean Series marking events where none of
    D1, D2, or fsc_small are > 1.

    Parameters
    ----------
    numpy.ndarray
        Boolean array of noise events.
    """
    if len(set(list(df)).intersection(set(["D1", "D2", "fsc_small"]))) < 3:
        raise ValueError("Can't apply noise filter without D1, D2, and fsc_small")

    # Mark noise events in new column "noise"
    return ~((df["fsc_small"].values > 1) | (df["D1"].values > 1) | (df["D2"].values > 1))


def mark_saturated(df):
    """
    Mark data that saturates D1 or D2.

    This function returns a boolean Series marking events where none of
    D1 == max(D1) or D2 == max(D2).

    Parameters
    ----------
    numpy.ndarray
        Boolean array of saturated events.
    """
    if len(set(list(df)).intersection(set(["D1", "D2"]))) < 2:
        raise ValueError("Can't apply saturation filter without D1 and D2")
    if len(df.index) == 0:
        return np.full(len(df.index), False)
    else:
        return (df["D1"].values == df["D1"].values.max()) | (df["D2"].values == df["D2"].values.max())


def merge_opp_vct(oppdf, vctdf):
    """
    Return a new DataFrame that combines an OPP DataFrame with a VCT DataFrame
    for one quantile.

    Parameters
    ----------
    oppdf: pandas.DataFrame
        SeaFlow OPP data for one quantile.
    vctdf: pandas.DataFrame
        SeaFlow VCT data for one quantile.

    Returns
    -------
    pandas.DataFrame
        New merged DataFrame.
    """
    if len(oppdf) != len(vctdf):
        raise ValueError("oppdf and vctdf must have the same number of rows")
    # Concat will join data by row index, so make sure both dataframes have the
    # same sequential indexes here first.
    oppdf = oppdf.reset_index(drop=True)
    vctdf = vctdf.reset_index(drop=True)
    return pd.concat([oppdf, vctdf], axis=1)


def quantiles_in_df(df):
    """
    Generator to iterate through focused particles by quantile.

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow particle data with focused particles marked by mark_focused().
        Focused particles should be marked with a boolean column for each
        quantile, where column names are q<quantile>, e.g. q2.5 for 2.5%
        quantile.

    Yields
    ------
    q_col: str
        Name of a single quantile focused boolean column.
    q: float
        Quantile number.
    q_str: str
        String representation of quantile suitable for constructing a filesystem
        path.
    q_df: pandas.DataFrame
        Subset of input DataFrame with only particles marked for the quantile
        defined by q_str.
    """
    for q_col in [c for c in df.columns if c.startswith("q")]:
        q_str = util.quantile_str(float(q_col[1:]))  # after "q"
        q = float(q_str)
        q_df = df[df[q_col]]  # select only focused particles for one quantile
        yield q_col, q, q_str, q_df


def roughfilter(df, width=5000):
    """
    Filter EVT particle data without bead positions or instrument calibration.

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow event data.
    width: int

    Returns
    -------
    pandas.DataFrame
        Copy of subset of df where each row is focused in at least on quantile.
    """
    if width is None:
        raise ValueError("Must supply width to roughfilter")
    # Prevent potential integer division bugs
    width = float(width)

    if len(df) == 0:
        return df

    # Mark noise and saturated particles
    noise = mark_noise(df)
    sat = mark_saturated(df)

    if np.sum(~noise & ~sat) == 0:
        # All data is noise/saturation filtered
        return df[0:0].copy()

    # Correction for the difference in sensitivity between D1 and D2
    origin = (df["D2"] - df["D1"]).median()

    # Filter aligned particles (D1 = D2), with correction for D1 D2
    # sensitivity difference.
    alignedD1 = (df["D1"] + origin) < (df["D2"] + width)
    alignedD2 = df["D2"] < (df["D1"] + origin + width)
    aligned = df[~noise & ~sat & alignedD1 & alignedD2]

    # Find fsc/d ratio (slope) for best large fsc particle
    fsc_small_max = aligned["fsc_small"].max()
    # Smallest D1 with maximum fsc_small
    min_d1 = aligned[aligned["fsc_small"] == fsc_small_max]["D1"].min()
    slope_d1 = fsc_small_max / min_d1
    # Smallest D2 with maximum fsc_small
    min_d2 = aligned[aligned["fsc_small"] == fsc_small_max]["D2"].min()
    slope_d2 = fsc_small_max / min_d2

    # Filter focused particles
    # Better fsc/d signal than best large fsc particle
    oppD1 = (aligned["fsc_small"] / aligned["D1"]) >= slope_d1
    oppD2 = (aligned["fsc_small"] / aligned["D2"]) >= slope_d2
    oppdf = aligned[oppD1 & oppD2].copy()

    return oppdf


def select_focused(df):
    """
    Return a DataFrame with particles that are focused at least one quantile.

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow event data that has been marked with mark_focused().

    Returns
    -------
    pandas.DataFrame
        Copy of subset of df where each row is focused in at least on quantile.
    """
    selector = False
    for qcolumn in [c for c in df.columns if c.startswith("q")]:
        selector = selector | df[qcolumn].values
    return df[selector].copy()


def linearize_particles(df, columns):
    """
    Linearize logged SeaFlow data.

    SeaFlow data is stored as log values over 3.5 decades on a 16-bit linear
    scale. This functions exponentiates those values onto a linear scale from 1
    to 10**3.5

    Note: This will convert to float64 if necessary.

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow event data.
    columns: list of str
        Names of columns to linearize.

    Returns
    -------
    pandas.DataFrame
        Copy of df with linearized values.
    """
    events = df.copy()
    if len(events.index) > 0:
        events[columns] = 10**((events[columns] / 2**16) * 3.5)
    return events


def log_particles(df, columns):
    """
    Opposite of linearize_particles().

    Parameters
    ----------
    df: pandas.DataFrame
        SeaFlow event data.
    columns: list of str
        Names of columns to log.

    Returns
    -------
    pandas.DataFrame
        Copy of df with logged values.
    """
    events = df.copy()
    if len(events.index) > 0:
        events[columns] = (np.log10(events[columns]) / 3.5) * 2**16
        events[columns] = events[columns].round(0)
    return events



"""Do things to SFL data DataFrames"""
from collections import OrderedDict
import json
import os
import re
import numpy as np
import pandas as pd
import pytz
from . import db
from . import errors as sfperrors
from . import geo
from . import time
from . import util
from . import seaflowfile


sfl_delim = '\t'

# Mappings between SFL file and SQL table column names
colname_mapping = {
    "table_to_file": {
        "file": "FILE",
        "date": "DATE",
        "file_duration": "FILE DURATION",
        "lat": "LAT",
        "lon": "LON",
        "conductivity": "CONDUCTIVITY",
        "salinity": "SALINITY",
        "ocean_tmp": "OCEAN TEMP",  # not a typo
        "par": "PAR",
        "bulk_red": "BULK RED",
        "stream_pressure": "STREAM PRESSURE",
        "flow_rate": "FLOW RATE",
        "event_rate": "EVENT RATE"
    }
}
# Reverse mappings from SQL table to file column headers
colname_mapping["file_to_table"] = {v:k for k, v in colname_mapping["table_to_file"].items()}

# Numeric columns using SQL table column names
numeric_columns = [
    "file_duration", "lat", "lon", "conductivity", "salinity", "ocean_tmp", "par", "bulk_red",
    "stream_pressure", "event_rate"
]

output_columns = [
    "file", "date", "file_duration", "lat", "lon", "conductivity",
    "salinity", "ocean_tmp", "par", "bulk_red", "stream_pressure",
    "event_rate"
]

# numeric min/max values
min_file_duration = 0
min_lat, max_lat = -90, 90
min_lon, max_lon = -180, 180
min_stream_pressure = 1e-4
min_event_rate = 0


def add_date_column(df):
    """Add a date column if needed and return a new dataframe."""
    newdf = df.copy(deep=True)

    def date_from_file(f):
        try:
            d = seaflowfile.SeaFlowFile(f).rfc3339
        except sfperrors.FileError:
            d = ''
        return d

    if "date" not in newdf.columns:
        newdf["date"] = newdf["file"].map(date_from_file)
    return newdf


def check(df):
    """Perform checks on SFL dataframe

    Returns a list of errors.
    """
    errors = []
    errors.extend(check_numeric(df, "file_duration", minval=min_file_duration, require_all=True))
    errors.extend(check_numeric(df, "lat", minval=min_lat, maxval=max_lat, require_all=False, require_some=True, warn_missing=True))
    errors.extend(check_numeric(df, "lon", minval=min_lon, maxval=max_lon, require_all=False, require_some=True, warn_missing=True))
    errors.extend(check_numeric(df, "conductivity", require_all=False, warn_missing=True))
    errors.extend(check_numeric(df, "salinity", require_all=False, warn_missing=True))
    errors.extend(check_numeric(df, "ocean_tmp", require_all=False, warn_missing=True))
    errors.extend(check_numeric(df, "par", require_all=False, warn_missing=True))
    errors.extend(check_numeric(df, "bulk_red", require_all=False, warn_missing=True))
    errors.extend(check_numeric(df, "stream_pressure", minval=min_stream_pressure, require_all=True))
    errors.extend(check_numeric(df, "event_rate", minval=min_event_rate, require_all=True))
    errors.extend(check_file(df))
    errors.extend(check_date(df))
    return errors


def check_date(df):
    errors = []
    if "date" not in df.columns:
        errors.append(create_error(df, "date", msg="date column is missing"))
    else:
        # All dates must match RFC 3339 with no fractional seconds
        # only integer seconds.
        date_flags = df["date"].map(check_date_string)
        if len(date_flags) > 0:
            # select rows that failed date check
            for i, v in df[~date_flags]["date"].iteritems():
                errors.append(create_error(df, "date", msg="Invalid date format", row=i, val=v))
    return errors


def check_date_string(date):
    """Confirm value is an RFC3339 string with UTC timezone as [+-]00:00"""
    passed = False
    try:
        dt = time.parse_date(date)
    except ValueError:
        pass
    else:
        if dt.tzinfo == pytz.UTC and (date.endswith('+00:00') or date.endswith('-00:00')):
            passed = True
    # Return true if any format is correct
    return passed


def check_file(df):
    errors = []
    # File field must be present
    if "file" not in df.columns:
        errors.append(create_error(df, "file", msg="file column is missing"))
    else:
        # File field must contain well formatted file strings, valid dates, and
        # day-of-year directory.
        def parse_filename(f):
            try:
                s = seaflowfile.SeaFlowFile(f)
            except sfperrors.FileError:
                return False
            if s.path_dayofyear:
                return True
            return False

        good_files_selector = df["file"].map(parse_filename)
        good_files = df[good_files_selector]
        bad_files = df[~good_files_selector]
        for i, v in bad_files["file"].iteritems():
            errors.append(create_error(bad_files, "file", msg="Invalid file name", row=i, val=v))

        # Files must be unique
        dup_files = df.loc[df["file"].duplicated(keep=False), "file"]
        for i, v in dup_files.iteritems():
            errors.append(create_error(dup_files, "file", msg="Duplicate file", row=i, val=v))

        # Files should be in order
        # Only consider files that are parseable by seaflowfile.SeaFlowFile
        inorder = seaflowfile.sorted_files(good_files["file"])
        files_equal = good_files["file"] == inorder
        if not files_equal.all():
            i = int(good_files[~files_equal].index[0])
            v = "First out of order file {}".format(good_files.loc[i, "file"])
            errors.append(create_error(good_files, "file", msg="Files out of order", row=i, val=v))

        # Files should match date in same row
        if "date" in df.columns:
            for i, v in good_files["file"].iteritems():
                s = seaflowfile.SeaFlowFile(v)
                d = good_files.loc[i, "date"]
                if s.is_new_style and s.rfc3339 != d:
                    errors.append(create_error(good_files, "file/date", msg="File and date don't match", row=i, val=f"{v} {d}"))

    return errors


def check_numeric(df, colname, require_all=False, require_some=False, warn_missing=False, minval=None, maxval=None):
    """
    Check a numeric column.

    Checks that the column is present and has valid numeric values. Optionally
    checks if values are in acceptable range and if all rows have non-NA values.

    Parameters
    -----------
    df: pandas.DataFrame
        SFL DataFrame with column to check.
    colname: str
        Name of the column to check.
    require_all: bool, default False
        Require that all rows have non-NA value?
    require_some: bool, default False
        Require that > 0 rows have non-NA value?
    warning_missing: bool, default False
        Warn when any rows have missing data?
    minval: int or float, optional
        Minimum acceptable value.
    maxval: int or float, optional
        Maximum acceptable value.

    Returns
    -------
    list of error dicts
        Errors created with create_error()
    """
    errors = []
    if colname not in df.columns:
        # column must be present
        errors.append(create_error(df, colname, msg=f"{colname} column is missing", level="error"))
    else:
        notnas = df[df[colname].notna()]
        numbers = pd.to_numeric(notnas[colname], errors="coerce")
        # Create boolean index for values in acceptable range
        # Start by selecting everything, then select by minval/maxval
        good_selector = np.ones(len(numbers), dtype=bool)
        if minval is not None:
            good_selector = good_selector & (numbers >= minval)
        if maxval is not None:
            good_selector = good_selector & (numbers <= maxval)
        # Catch values outside correct range
        # Catch non-numeric values (NAs created during to_numeric())
        bad_numbers = notnas.loc[~good_selector, colname]
        for i, v in bad_numbers.iteritems():
            errors.append(create_error(df, colname, msg=f"Invalid {colname}", row=i, val=v, level="error"))

        nas = df.loc[df[colname].isna(), colname]
        if len(nas) == len(df):
            # No data in column
            if require_all or require_some:
                errors.append(create_error(df, colname, msg=f"{colname} column has no data", level="error"))
            else:
                errors.append(create_error(df, colname, msg=f"{colname} column has no data", level="warning"))
        elif len(nas) > 0:
            # Some missing
            if require_all:
                for i, v in nas.iteritems():
                    errors.append(create_error(df, colname, msg="Missing required data", row=i, val=v, level="error"))
            elif warn_missing:
                for i, v in nas.iteritems():
                    errors.append(create_error(df, colname, msg="Missing data", row=i, val=v, level="warning"))

    return errors


def convert_gga2dd(df):
    """Return a copy of df with coordinates converted from GGA to decimal degrees."""
    newdf = df.copy(deep=True)
    newdf["lat"] = df["lat"].map(geo.ggalat2dd, na_action="ignore")
    newdf["lon"] = df["lon"].map(geo.ggalon2dd, na_action="ignore")
    return newdf


def create_error(df, col, msg, row=None, val=None, level='error'):
    """Create an error dictionary.

    Error levels can be 'error' (fatal) or 'warning'.
    """
    level_values = ['error', 'warning']
    if level not in level_values:
        raise ValueError(f"valid values for 'level' are {level_values}")

    e = {
        "column": col,
        "message": msg,
        "line (1-based)": None,
        "value": None,
        "level": level,
    }
    if row is not None:
        e["line (1-based)"] = row + 2
    if val is not None:
        e["value"] = make_json_serializable(val)
    elif row is not None:
        e["value"] = make_json_serializable(df.loc[row, col])
    return e


def dedup(df):
    """Remove duplicate rows from DataFrame by "file".

    Returns a 2-tuple of:
        - unique duplicate files, also as 2-tuple of (file_id, count)
        - df without duplicate file rows
    """
    # Find all duplicate files
    dups = df[df.duplicated("file", keep=False)]["file"].tolist()
    # Create a unique list of duplicate file names in order of occurrence
    d = OrderedDict()
    for f in dups:
        if f not in d:
            d[f] = 0
        d[f] += 1
    return (list(d.items()), df.drop_duplicates(subset="file", keep=False))


def find_sfl_files(root):
    """Find all files with .sfl extension beneath root.

    Returns a list of sfl file paths relative to root.
    """
    root = os.path.expanduser(root)
    sfl_paths = []
    for dirpath, _dirnames, filenames in os.walk(root):
        for f in filenames:
            if f.endswith(".sfl"):
                sfl_paths.append(os.path.join(dirpath, f))
    return sorted(sfl_paths)


def fix(df):
    """Return a copy of df ready for db import.

    - Adds a "date" column if not present, based on filename datestamp
    - Adds or replaces day of year directory component of "file" values
    - Set any stream pressure values <= 0 to 1e-4 (small positive number)
    - Adds any missing db columns
    """
    newdf = df.copy(deep=True)

    newdf = add_date_column(newdf)

    # Add day of year directory if needed
    def dayofyear_from_file(f):
        try:
            d = seaflowfile.SeaFlowFile(f).file_id
        except sfperrors.FileError:
            d = f  # don't change anything if can't parse filename
        return d

    newdf["file"] = newdf["file"].map(dayofyear_from_file)

    # Convert stream pressure <= 0 to small positive number
    newdf.loc[newdf["stream_pressure"] <= 0, "stream_pressure"] = min_stream_pressure

    # Make sure all DB columns are present
    for k in colname_mapping["table_to_file"]:
        if k not in newdf.columns:
            newdf[k] = None

    return newdf


def fix_event_rate(df, event_counts):
    """
    Update event_rate field based on event counts in event_counts.

    Parameters
    -----------
    df: pandas DataFrame
        SFL DataFrame, based on a "fixed" file
    event_counts: dict of {str: int}
        Dictionary with file_id: event count

    Returns
    -------
    df: pandas DataFrame
        Copy of df with updated event_rate fields where possible.
    """
    newdf = df.copy(deep=True)
    for i, row in newdf.iterrows():
        try:
            file_duration = float(row["file_duration"])
        except ValueError:
            continue
        try:
            event_count = int(event_counts[row["file"]])
        except (ValueError, KeyError):
            continue
        try:
            newdf.loc[i, "event_rate"] = event_count / file_duration
        except ZeroDivisionError:
            newdf.iloc[i, "event_rate"] = 0.0
    return newdf


def fix_underway(sfl_df, geo_df, thermo_df):
    """
    Update sfl_df underway columns based on data in geo_df and thermo_df. geo_df
    and thermo_df will be resampled down to 1 minute time frequency before
    merging into sfl_df by matching datetime to the minute.

    Parameters
    -----------
    sfl_df: pandas DataFrame
        SFL DataFrame, based on a "fixed" file
    geo_df: pandas DataFrame
        Should have columns [datetime "time", float "lat", "lon"]
    thermo_df: pandas DataFrame
        Should have columns [datetime "time", float "ocean_tmp", "salinity", "conductivity"]

    Returns
    -------
    df: pandas DataFrame
        Copy of sfl_df with updated underway columns.
    """
    # Copy everything first
    sfl_df, geo_df, thermo_df = sfl_df.copy(), geo_df.copy(), thermo_df.copy()

    offset = "1min"  # resampling target frequency
    # Resample geo
    # Special case handling of longitude around the international date line
    idx = geo_df["lon"] <= -175 # assuming a ship won't go 5 degrees in 1 minute
    geo_df.loc[idx, "lon"] = geo_df.loc[idx, "lon"] + 360
    geo_df2 = geo_df.resample(offset, on="time").mean()
    idx = geo_df2["lon"] > 180
    geo_df2.loc[idx, "lon"] = geo_df2.loc[idx, "lon"] - 360
    # Resample thermo
    thermo_df2 = thermo_df.resample(offset, on="time").mean()

    # Merge into SFL by matching minutes
    sfl_df["time"] = pd.to_datetime(sfl_df["date"]).dt.floor("1min") # remove later
    merged = sfl_df.merge(geo_df2, how="left", on="time")
    merged = merged.merge(thermo_df2, how="left", on="time")
    for col in ["lat", "lon", "ocean_tmp", "salinity", "conductivity"]:
        notna = merged[col + "_y"].notna()
        print("merging {:d} rows".format(len(notna)))
        merged.loc[notna, col + "_x"] = merged.loc[notna, col + "_y"]
    # Remove temporary "time" column
    merged = merged.drop(["time"], axis=1)
    # Remove columns from other dataframes
    to_drop = [col for col in merged.columns if col.endswith("_y")]
    merged = merged.drop(to_drop, axis=1)
    # Remove _x from left merged columns
    to_rename = {col: col.rstrip("_x") for col in merged.columns if col.endswith("_x")}
    merged = merged.rename(to_rename, axis=1)

    return merged

def has_gga(df):
    """Do any coordinates Series in this DataFrame contain GGA values?"""
    gga_lats = df["lat"].map(geo.is_gga_lat, na_action="ignore")
    gga_lons = df["lon"].map(geo.is_gga_lon, na_action="ignore")
    return (gga_lats | gga_lons).any()


def make_json_serializable(v):
    """Make sure v is JSON serializable if it's numpy type or plain object"""
    if isinstance(v, np.generic):
        return np.asscalar(v)
    return v


def parse_sfl_filename(fn):
    fn = os.path.basename(fn)
    m = re.match(r"^(?P<cruise>.+)_(?P<inst>[^_]+).sfl$", fn)
    if m:
        return (m.group('cruise'), m.group('inst'))
    return ()


@util.suppress_sigpipe
def print_json_errors(errors, fh, filename, print_all=True):
    errors_output = []
    errors_seen = set()
    for e in errors:
        e["file"] = filename
        error_key = e["column"] + e["message"]
        if (not print_all) and (error_key in errors_seen):
            continue
        errors_seen.add(error_key)
        errors_output.append(e)
    fh.write(json.dumps(errors_output, sort_keys=True, indent=2, separators=(',', ':')))
    fh.write("\n")


@util.suppress_sigpipe
def print_tsv_errors(errors, fh, filename, print_all=True, header=True,):
    errors_output = []
    errors_seen = set()
    for e in errors:
        e["file"] = filename
        error_key = e["column"] + e["message"]
        if (not print_all) and (error_key in errors_seen):
            continue
        errors_seen.add(error_key)
        errors_output.append(e)

    # TSV output
    if header:
        header_written = False
    else:
        header_written = True
    for e in errors_output:
        if not header_written:
            print("\t".join([k for k in sorted(e.keys())]), file=fh)
            header_written = True
        print("\t".join([str(e[k]) for k in sorted(e.keys())]), file=fh)


def read_file(
        file_path, convert_dates=False, convert_numerics=True,
        convert_colnames=True, **kwargs
    ):
    """Parse SFL file into a DataFrame.

    Arguments:
    file -- SFL file path.

    Keyword arguments:
    convert_dates -- Convert date column to datetime objects (default False).
    convert_numerics -- Cast numeric SQL columns as numbers (default True).
    convert_colnames -- Remap file column names to match SFL SQL table column
        where appropriate. (default True).
    """
    defaults = {
        "sep": str(sfl_delim),
        "dtype": str,
        "na_filter": True,
        "encoding": "utf-8"
    }
    kwargs_defaults = dict(defaults, **kwargs)

    try:
        df = pd.read_csv(file_path, **kwargs_defaults)
    except pd.errors.ParserError:
        raise sfperrors.FileError("could not parse {} as an sfl file".format(file_path))
    df = df.rename(columns=colname_mapping["file_to_table"])

    if convert_dates:
        df = add_date_column(df)
        df["date"] = df["date"].map(time.parse_date)

    if convert_numerics:
        for colname in numeric_columns:
            df[colname] = pd.to_numeric(df[colname], errors='coerce')

    if not convert_colnames:
        # Revert column name mapping back to file convention
        df = df.rename(columns=colname_mapping["table_to_file"])

    return df


def save_to_db(df, dbpath, cruise=None, serial=None):
    """Write SFL dataframe to a SQLite3 database.

    Any pre-existing SFL data will be erased.

    Arguments:
    df -- SFL DataFrame.
    dbpath -- Path to SQLite3 database file.
    """
    db.create_db(dbpath)  # create or update db if needed
    if cruise is None:
        cruise = 'None'
    if serial is None:
        serial = 'None'
    metadf = pd.DataFrame({'cruise': [cruise], 'inst': [serial]})
    db.save_metadata(dbpath, metadf.to_dict('index').values())
    # This assumes there are column names which match SQL SFL table
    db.save_sfl(dbpath, df.to_dict('index').values())


@util.suppress_sigpipe
def save_to_file(df, outpath, convert_colnames=True, all_columns=False):
    """Write SFL dataframe to a csv file.

    Arguments:
    df -- SFL DataFrame.
    outpath -- Output file path.

    Keyword Arguments:
    convert_colnames -- Remap SQL table column names to SFL file column names
        where appropriate. (default True).
    """
    # Remove input file path and line number columns that may have been
    # added.
    if not all_columns:
        df = df[output_columns]
    if convert_colnames:
        df = df.rename(columns=colname_mapping["table_to_file"])
    df.to_csv(outpath, sep=str(sfl_delim), na_rep="NA", encoding="utf-8",
        index=False, float_format="%.4f")



class SeaFlowpyError(Exception):
    """Base class for all seaflowpy exceptions"""
    pass


class FileError(SeaFlowpyError):
    """Custom exception class for SeaFlow file format errors"""
    pass


class ClusterError(SeaFlowpyError):
    """Custom exception class bead clustering errors"""
    pass


class S3Error(SeaFlowpyError):
    """Custom exception class for boto3 S3 related erorrs"""
    pass


import os
import re

import pandas as pd
from . import errors
from . import time
from . import util


dayofyear_re = r'^\d{1,4}_\d{1,3}$'
new_file_re = r'^(?P<date>\d{4}-\d{2}-\d{2})T(?P<hours>\d{2})-(?P<minutes>\d{2})-(?P<seconds>\d{2})(?P<tzhours>[+-]\d{2})-(?P<tzminutes>\d{2})$'
old_file_re = r'^\d+\.evt$'


class SeaFlowFile:
    """Base class for EVT file classes"""

    def __init__(self, path, date=None):
        """
        The date can be set from the date argument for old style filenames that
        don't have timestamps. For new style file names if the date parsed from
        the filename doesn't match the date passed in the contructor a ValueError
        will be raised.
        """
        self.path = path

        parts = parse_path(self.path)
        self.filename = parts["file"]
        if parts["file"].endswith(".gz"):
            self.filename_nogz = self.filename[:-3]
        else:
            self.filename_nogz = self.filename

        if not (self.is_old_style or self.is_new_style):
            raise errors.FileError("Filename doesn't look like a SeaFlow EVT file")

        if self.is_new_style:
            err = None
            try:
                timestamp = timestamp_from_filename(self.filename_nogz)
                self.date = time.parse_date(timestamp)
            except ValueError as e:
                err = e
            if err:
                raise errors.FileError(str(err))
            if date is not None and self.date != date:
                raise ValueError(
                    "parsed date does not match date argument, {} != {}".format(
                        self.date.isoformat(timespec="seconds"),
                        date.isoformat(timespec="seconds")
                    )
                )
        else:
            self.date = date

        # YYYY_dayofyear directory found in file path and parsed
        # from file datestmap
        self.path_dayofyear = parts["dayofyear"]

        # Identifer to match EVT/SFL files
        # Should be something like 2014_142/42.evt for old files.
        # Should be something like 2014_342/2014-12-08T22-53-34+00-00 for new
        # files. Note no extension including .gz.
        # The day of year directory will be based on parsed datestamp in
        # filename when possible, not the given path. The file ID based on
        # the given path is stored in path_file_id.
        if self.is_old_style:
            # path_file_id and file_id are always the same for old-style
            # filenames since we can't parse dates to calculate a day of year
            # directory
            if self.path_dayofyear:
                self.file_id = "{}/{}".format(self.path_dayofyear, self.filename_nogz)
            else:
                self.file_id = self.filename_nogz
            self.path_file_id = self.file_id
        else:
            self.file_id = "{}/{}".format(self.dayofyear, self.filename_nogz)
            if self.path_dayofyear:
                self.path_file_id = "{}/{}".format(self.path_dayofyear, self.filename_nogz)
            else:
                self.path_file_id = self.filename_nogz

    def __str__(self):
        return "SeaFlowFile: {}, {}".format(self.file_id, self.path)

    @property
    def dayofyear(self):
        """Return day of year based on date."""
        return create_dayofyear_directory(self.date)

    @property
    def isgz(self):
        """Is file gzipped?"""
        return self.path and self.filename.endswith(".gz")

    @property
    def is_old_style(self):
        """Is this old style file? e.g. 2014_185/1.evt."""
        return bool(re.match(old_file_re, self.filename_nogz))

    @property
    def is_new_style(self):
        """Is this a new style file? e.g. 2018_082/2018-03-23T00-00-00+00-00.gz"""
        return bool(re.match(new_file_re, self.filename_nogz))

    @property
    def rfc3339(self):
        """Return RFC 3339 YYYY-MM-DDThh:mm:ss[+-]hh:mm parsed from filename"""
        if self.date:
            return self.date.isoformat(timespec='seconds')
        return ''

    @property
    def sort_key(self):
        # day of year from filename if possible first, then from path, then nothing
        if self.dayofyear:
            year, day = [int(x) for x in self.dayofyear.split("_")]
        elif self.path_dayofyear:
            year, day = [int(x) for x in self.path_dayofyear.split("_")]
        else:
            year, day = 0, 0
        if self.is_old_style:
            # Number part of basename, necessary because number isn't
            # zero-filled
            file_key = int(self.filename_nogz.split(".")[0])
        else:
            file_key = self.filename_nogz
        return (year, day, file_key)


def create_dayofyear_directory(dt):
    """Create SeaFlow day of year directory from a datetime object"""
    if dt:
        return "{}_{}".format(dt.year, dt.strftime('%j'))
    return ''


def timestamp_from_filename(filename):
    filename_to_first_dot = os.path.basename(filename).split(".")[0]
    m = re.match(new_file_re, filename_to_first_dot)
    if m:
        # New style EVT/SFL filenames, e.g.
        # - 2014-05-15T17-07-08+00-00
        # - 2014-05-15T17-07-08-07-00
        # Parse RFC 3339 date string
        return "{date}T{hours}:{minutes}:{seconds}{tzhours}:{tzminutes}".format(**m.groupdict())
    raise ValueError('filename does not look like a new-style SeaFlow file')


def parse_path(file_path):
    """Return a dict with entries for 'dayofyear' dir and 'file' name"""
    d = {"dayofyear": '', "file": ''}
    parts = util.splitpath(file_path)
    d["file"] = parts[-1]
    if len(parts) > 1:
        if re.match(dayofyear_re, parts[-2]):
            d["dayofyear"] = parts[-2]
    return d


def sorted_files(files):
    """Sort EVT file paths in chronological order.

    Order is based on day of year directory parsed from path and then file name.
    """
    sfiles = [SeaFlowFile(f) for f in files]
    return [s.path for s in sorted(sfiles, key=lambda x: x.sort_key)]


def filtered_file_list(total_list, filter_list):
    """
    Only keep files from total_list that are in filter_list.

    Match by file_id, but return original path in total_list.
    """
    filter_set = {SeaFlowFile(f).file_id for f in filter_list}
    files = []
    for f in total_list:
        if SeaFlowFile(f).file_id in filter_set:
            files.append(f)
    files = sorted_files(files)
    return files


def find_evt_files(root_dir):
    """Return a chronologically sorted list of EVT file paths in root_dir."""
    files = util.find_files(root_dir)
    files = keep_evt_files(files)
    return sorted_files(files)


def keep_evt_files(files):
    """Filter list of files to only keep EVT files."""
    files_list = []
    for f in files:
        try:
            _ = SeaFlowFile(f)
        except errors.FileError:
            pass
        else:
            files_list.append(f)
    return files_list


def timeselect_evt_files(sfiles, tstart, tend):
    """
    Filter a list of EVT files by datetime.datetime start and end.

    Parameters
    -----------
    sfiles: iterable of seaflowfile.SeaFlowFile
        Any files without dates will not be selected.
    tstart: str
        Start datetime. Pass None to remove lower bound.
    tend: str
        End datetime. Pass None to remove upper bound.

    Returns
    -------
    List of seaflowfile.SeaFlowFile within tstart and tend.
    """
    sfiles = [f for f in sfiles if f.date is not None]
    if tstart is not None:
        sfiles = [f for f in sfiles if f.date >= tstart]
    if tend is not None:
        sfiles = [f for f in sfiles if f.date <= tend]
    return sfiles


def date_evt_files(evt_paths, sfl_df=None):
    """
    Create a DataFrame of file IDs, paths, timestamps.

    Parameters
    -----------
    evt_paths: list of str
        EVT file paths.
    sfl_df: pandas.DataFrame
        DataFrame for SFL data, with "file" column for file IDs and "date" column
        with RFC3339 timestamp strings or datetime objects. If sfl_df is
        provided, only files in evt_paths and sfl_df["file"] that share file IDs
        will be included. If it is not provided dates will be parsed from file
        names if possible.

    Raises
    ------
    ValueError if dates in SFL can't be parsed.
    KeyError if "file" or "date" is missing from sfl_df.
    seaflowpy.errors.SeaFlowpyError if a filename can't be parsed.

    Returns
    -------
    pandas.DataFrame
        "file_id" column has file IDs, "path" has file paths, "date" has
        timestamp objects.
    """
    data = {"date": [], "file_id": [], "path": []}
    if sfl_df is not None:
        if not pd.api.types.is_datetime64_ns_dtype(sfl_df["date"]):
            sfl_df["date"] = sfl_df["date"].map(time.parse_date)
        sfl_dates_by_file = dict(zip(sfl_df["file"].tolist(), sfl_df["date"].tolist()))
        for path in evt_paths:
            file_id = SeaFlowFile(path).file_id
            if file_id in sfl_dates_by_file:
                data["file_id"].append(file_id)
                data["path"].append(path)
                data["date"].append(sfl_dates_by_file[file_id])
    else:
        for path in evt_paths:
            sff = SeaFlowFile(path)
            data["file_id"].append(sff.file_id)
            data["path"].append(path)
            data["date"].append(sff.date)
    return pd.DataFrame(data)[["date", "file_id", "path"]]



import sys
import click
from seaflowpy import db
from seaflowpy.errors import SeaFlowpyError
from seaflowpy import fileio
from seaflowpy import sfl

# Subcommand aliases for backwards compatibility
aliases = {'create': 'import-sfl'}


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        try:
            real_cmd = aliases[cmd_name]
            return click.Group.get_command(self, ctx, real_cmd)
        except KeyError:
            return None


@click.command(cls=AliasedGroup)
def db_cmd():
    """Database file operations subcommand."""
    pass


@db_cmd.command('create')
@click.option('-d', '--db', 'db_file', required=True, type=click.Path(writable=True),
    help='DB file to create or update.')
@click.option('-c', '--cruise', required=True,
    help='Supply a cruise name to be saved in the database.',)
@click.option('-s', '--serial', required=True,
    help='Supply an instrument serial number to be saved in the database.')
def db_create_cmd(db_file, cruise, serial):
    """
    Create or updated a popcycle database with cruise and serial number.
    """
    try:
        db.save_metadata(db_file, [{"cruise": cruise, "inst": serial}])
    except SeaFlowpyError as e:
        raise click.ClickException(str(e))


@db_cmd.command('import-sfl')
@click.option('-f', '--force', is_flag=True,
    help='Attempt DB import even if validation produces errors.')
@click.option('-j', '--json', is_flag=True,
    help='Report errors as JSON.')
@click.option('-v', '--verbose', is_flag=True,
    help='Report all errors.')
@click.argument('sfl-file', nargs=1, type=click.Path(exists=True))
@click.argument('db-file', nargs=1, type=click.Path(writable=True))
def db_import_sfl_cmd(force, json, verbose, sfl_file, db_file):
    """
    Imports SFL metadata to database.

    Writes processed SFL-FILE data to SQLite3 database file. Data will be
    checked before inserting. If any errors are found the first of each type
    will be reported and no data will be written. SFL-FILE may have the 
    <cruise name> and <instrument serial>
    embedded in the filename as '<cruise name>_<instrument serial>.sfl'. If not,
    it's expected that information is already in the database. Databae cruise and
    serial overrides values in filename. If a database
    file does not exist a new one will be created. Any SFL data in the database
    will be erased before importing new data. Errors or warnings are output to
    STDOUT.
    """
    cruise, serial = None, None

    # Try to read cruise and serial from database
    try:
        cruise = db.get_cruise(db_file)
    except SeaFlowpyError as e:
        pass
    try:
        serial = db.get_serial(db_file)
    except SeaFlowpyError as e:
        pass

    # Try to read cruise and serial from filename if not already defined
    if cruise is None or serial is None:
        results = sfl.parse_sfl_filename(sfl_file)
        if results:
            cruise = results[0]
            serial = results[1]

    # Make sure cruise and serial are defined somewhere
    if cruise is None or serial is None:
        raise click.ClickException('instrument serial and cruise must both be specified either in filename as <cruise>_<instrument-serial>.sfl, as command-line options, or in database metadata table.')

    df = sfl.read_file(sfl_file)

    df = sfl.fix(df)
    errors = sfl.check(df)

    if len(errors) > 0:
        if json:
            sfl.print_json_errors(errors, sys.stdout, sfl_file, print_all=verbose)
        else:
            sfl.print_tsv_errors(errors, sys.stdout, sfl_file, print_all=verbose)
        if not force and len([e for e in errors if e["level"] == "error"]) > 0:
            sys.exit(1)
    sfl.save_to_db(df, db_file, cruise, serial)


@db_cmd.command('import-filter-params')
@click.option('-c', '--cruise',
    help='Supply a cruise name for parameter selection. If not provided cruise in database will be used.')
@click.argument('filter-file', nargs=1, type=click.File())
@click.argument('db-file', nargs=1, type=click.Path(writable=True))
def db_import_filter_params_cmd(cruise, filter_file, db_file):
    """
    Imports filter parameters to database.

    A new database will be created if it doesn't exist.
    """
    # If cruise not supplied, try to get from db
    if cruise is None:
        try:
            cruise = db.get_cruise(db_file)
            print("using cruise={} from db".format(cruise))
        except SeaFlowpyError:
            pass

    if cruise is None:
        raise click.ClickException('cruise must be specified either as command-line option or in database metadata table.')

    df = fileio.read_filter_params_csv(filter_file)
    params = df[df.cruise == cruise]
    if len(params.index) == 0:
        raise click.ClickException('no filter parameters found for cruise %s' % cruise)
    db.save_filter_params(db_file, params.to_dict('index').values())


@db_cmd.command('merge')
@click.argument('db1', type=click.Path(exists=True))
@click.argument('db2', type=click.Path(exists=True, writable=True))
def db_merge_cmd(db1, db2):
    """Merges SQLite3 DB1 into DB2.

    Only merges gating, poly, filter tables.
    """
    db.merge_dbs(db1, db2)



from contextlib import contextmanager
import gzip
import io
import os
import zlib
import numpy as np
import pandas as pd
from . import errors
from . import particleops
from .seaflowfile import SeaFlowFile
from . import util


@contextmanager
def file_open_r(path, fileobj=None, as_text=False):
    """
    Open path or fileobj for reading as a context manager.

    Data read from the return value of this function will come from path or
    preferentially fileobj if provided. If path is provided and ends with '.gz'
    data will be considered gzip compressed even if read from fileobj. All
    resources opened by this function (input file handles) or open resources
    passed to this function (fileobj) will be cleaned up by context managers.
    The return value of this function should always be used within a 'with'
    block.

    Parameters
    -----------
    path: str
        File path.
    fileobj: io.BytesIO, optional
        Open file object.
    as_text: bool, default False
        Return a file-like object of strings rather than bytes

    Returns
    -------
    Context manager for file-like object of bytes or strings.
    """
    # zlib is faster than gzip for decompression of EVT data on MacOS, and
    # comparable on Linux.
    if fileobj:
        if path.endswith('.gz'):
            gzdata = fileobj.read()
            zobj = zlib.decompressobj(wbits=zlib.MAX_WBITS|32)
            data = zobj.decompress(gzdata)
            if not as_text:
                yield io.BytesIO(data)
            else:
                yield io.TextIOWrapper(io.BytesIO(data))
        else:
            if not as_text:
                yield fileobj
            else:
                yield io.TextIOWrapper(fileobj)
    else:
        if path.endswith('.gz'):
            with io.open(path, 'rb') as fileobj:
                gzdata = fileobj.read()
                zobj = zlib.decompressobj(wbits=zlib.MAX_WBITS|32)
                data = zobj.decompress(gzdata)
                if not as_text:
                    yield io.BytesIO(data)
                else:
                    yield io.TextIOWrapper(io.BytesIO(data))
        else:
            if not as_text:
                with io.open(path, 'rb') as fh:
                    yield fh
            else:
                with io.open(path, 'r') as fh:
                    yield fh


@contextmanager
def file_open_w(path):
    """
    Open path for writing as a context manager.

    If path ends with '.gz' data will gzip compressed. Only the write method of
    the returned object should be used. All resources opened in this function
    (output file handles, gzipping child processes) will be cleaned up by
    context managers. The return value of this function should always be used
    within a 'with' block.

    Parameters
    -----------
    path: str
        File path.

    Returns
    -------
    Context manager for writable file-like object.
    """
    if path.endswith('.gz'):
        with gzip.open(path, mode='wb', compresslevel=9) as fh:
            yield fh
    else:
        with io.open(path, 'wb') as fh:
            yield fh


def read_evt_header(f):
    """
    Detect v1 or v2 EVT file.

    Parameters
    -----------
    f: Open file-like object

    Raises
    ------
    errors.FileError
        When version can't be determined
    Returns
    -------
    dict
        { "rowcnt": row count, "colcnt": column count, "version": v1 or v2 }
    """
    bytes_per_col = 2

    # v1: number of rows of data
    # v2: bytes per row of data, should always be 14 (7 columns * 2 bytes)
    buff1 = f.read(4)
    # v1: number of 2-byte columns per row, should always be 10
    # v2: number of rows of data
    buff2 = f.read(4)
    if len(buff1) < 4 or len(buff2) < 4:
        raise errors.FileError("File has incomplete leading 32bit numbers")
    num1 = int.from_bytes(buff1, byteorder='little')
    num2 = int.from_bytes(buff2, byteorder='little')
    if num2 == len(particleops.COLUMNS):
        return { "rowcnt": num1, "colcnt": num2, "version": "v1" }
    if num1 == len(particleops.COLUMNS2) * bytes_per_col:
        return { "rowcnt": num2, "colcnt": num1 / bytes_per_col, "version": "v2" }
    raise errors.FileError("File does not have a valid column size number")


def read_labview(path, columns=None, fileobj=None):
    """
    Read a labview binary SeaFlow data file.

    Data will be read from the file at the provided path or preferentially from
    fileobj if provided. If path is provided and ends with '.gz' data will be
    considered gzip compressed even if read from fileobj.

    Parameters
    -----------
    path: str
        File path.
    columns: list of str, default is None
        Names of columns. If none are given reasonable defaults are used.
    fileobj: io.BytesIO, optional
        Open file object.

    Returns
    -------
    dict of {
        "version": "v1"|"v2",
        "df": SeaFlow event pandas.DataFrame as numpy.float64 values
    }
    """
    try:
        with file_open_r(path, fileobj) as fh:
            counts = read_evt_header(fh)
            rowcnt, version = counts["rowcnt"], counts["version"]
            if rowcnt == 0:
                raise errors.FileError("File has no particle data")

            if version == "v1":
                # v1 EVT
                if columns is None:
                    columns = particleops.COLUMNS
                colcnt = len(columns)  # for compat with old labview OPP files
                # Each row has a leading 32-bit int (column count) which can be
                # discarded. We'll account for it by adding 2 to the colcnt
                # (2 extra 16-bit columns)
                colcnt += 2
                expected_bytes = rowcnt * colcnt * 2  # 2 bytes per column
                # Put the leading 32-bit int back in front of the first row
                # since we already read it to get colcnt.
                buff = int(colcnt).to_bytes(4, byteorder='little') + fh.read(int(expected_bytes) - 4)
            elif version == "v2":
                # v2 EVT
                if columns is None:
                    columns = particleops.COLUMNS2
                colcnt = len(columns)  # for compat with old labview OPP files
                # Unlike v1, there are no leading 32-bit ints for colcnt, except
                # for the one we read at the beginning.
                expected_bytes = rowcnt * colcnt * 2  # 2 bytes per column
                buff = fh.read(int(expected_bytes))
            else:
                raise ValueError("invalid version string")

            # Read any extra data at the end of the file for error checking. There
            # shouldn't be any extra data, btw.
            extra_bytes = 0
            while True:
                new_bytes = len(fh.read(8192))
                extra_bytes += new_bytes
                if new_bytes == 0:  # end of file
                    break
    except (IOError, EOFError, zlib.error) as e:
        raise errors.FileError("File could not be read: {}".format(str(e)))

    # Check that file has the expected number of data bytes.
    found_bytes = len(buff) + extra_bytes
    if found_bytes != expected_bytes:
        raise errors.FileError(
            "File has incorrect number of data bytes. Expected %i, saw %i" %
            (expected_bytes, found_bytes)
        )

    events = np.frombuffer(buff, dtype="uint16", count=rowcnt*colcnt)
    # Reshape into a matrix of colcnt columns and one row per particle
    events = np.reshape(events, [rowcnt, colcnt])
    # Create a Pandas DataFrame with descriptive column names.
    if version == "v1":
        # v1 file, remove leading two columns (32-bit column count int in each row)
        df = pd.DataFrame(np.delete(events, [0, 1], 1), columns=columns)
    else:
        df = pd.DataFrame(events, columns=columns)
    return {"version": version, "df": df.astype(np.float64)}


def read_evt_labview(path, fileobj=None):
    """
    Read a raw labview binary SeaFlow data file.

    Data will be read from the file at the provided path or preferentially from
    fileobj if provided. If path is provided and ends with '.gz' data will be
    considered gzip compressed even if read from fileobj.

    Parameters
    -----------
    path: str
        File path.
    fileobj: io.BytesIO, optional
        Open file object.

    Returns
    -------
    dict of {
        "version": "v1"|"v2",
        "df": SeaFlow event pandas.DataFrame as numpy.float64 values
    }
    """
    return read_labview(path, columns=None, fileobj=fileobj)


def read_filter_params_csv(path):
    """
    Read a filter parameters csv file.

    Parameters
    ----------
    path: str
        Path to filter parameters csv file.

    Returns
    -------
    pandas.DataFrame
        Contents of csv file with "." in column headers replaced with "_".
    """
    defaults = {
        "sep": str(','),
        "na_filter": True,
        "encoding": "utf-8"
    }
    try:
        df = pd.read_csv(path, **defaults)
    except pd.errors.ParserError:
        raise errors.FileError("could not parse {} as csv filter paramater file".format(path))
    # Fix column names
    df.columns = [c.replace('.', '_') for c in df.columns]
    # Make sure serial numbers are treated as strings
    # pandas.read_csv can return a dataframe or TextFileReader so disable
    # member method linting.
    # pylint: disable=no-member
    df = df.astype({"instrument": "str"})
    return df


def write_labview(df, path):
    """
    Write SeaFlow event DataFrame as LabView binary file.

    Parameters
    -----------
    df: pandas.DataFrame
        SeaFlow event DataFrame.
    path: str
        Output file path. If this ends with '.gz' data will be gzip compressed.
    """
    # Make sure directory necessary directory tree exists
    util.mkdir_p(os.path.dirname(path))

    # Open output file
    with file_open_w(path) as fh:
        # Write 32-bit uint particle count header
        header = np.array([len(df.index)], np.uint32)
        fh.write(header.tobytes())
        if len(df.index) > 0:
            # Convert to uint16 before saving
            df = df.astype(np.uint16)

            # Add leading 4 bytes to match LabViews binary format
            zeros = np.zeros([len(df.index), 1], dtype=np.uint16)
            tens = np.copy(zeros)
            tens.fill(10)
            df.insert(0, "tens", tens)
            df.insert(1, "zeros", zeros)

            # Write particle data
            fh.write(df.values.tobytes())


def write_evt_labview(df, path, outdir, gz=True):
    """
    Write a raw SeaFlow event DataFrame as LabView binary file.

    Parameters
    -----------
    df: pandas.DataFrame
        SeaFlow raw event DataFrame.
    path: str
        File name. This will be converted into a standard SeaFlow file ID and
        will be used to construct the final output file path within outdir. The
        final file name will a ".gz" extension if gz is True.
    outdir: str
        Output directory. This function will create day of year subdirectories
        for EVT binary files.
    gz: bool, default True
        Gzip compress?
    """
    if df is None:
        return

    sfile = SeaFlowFile(path)
    outpath = os.path.join(outdir, sfile.file_id)
    if gz:
        outpath = outpath + ".gz"
    # Only keep columns we intend to write to file
    write_labview(df[particleops.COLUMNS], outpath)


def write_opp_parquet(opp_dfs, date, window_size, outdir):
    """
    Write an OPP Parquet file.

    Use snappy compression.

    Parameters
    -----------
    opp_dfs: pandas.DataFrame
        SeaFlow focused particle DataFrames with file_id, date, and index reflecting
        positions in original EVT DataFrames.
    date: pandas.Timestamp or datetime.datetime object
        Start timestamp for data in df.
    window_size: pandas offset alias for time window covered by this file. Time
        covered by this file is date + time_window.
    outdir: str
        Output directory.
    """
    if not opp_dfs:
        return

    # Make sure directory necessary directory tree exists
    util.mkdir_p(outdir)
    outpath = os.path.join(outdir, date.isoformat().replace(":", "-")) + f".{window_size}.opp.parquet"
    df = pd.concat(opp_dfs, ignore_index=True)
    # Linearize data columns
    df = particleops.linearize_particles(df, columns=["D1", "D2", "fsc_small", "pe", "chl_small"])
    # Only keep columns we intend to write to file, reorder
    columns = [
        "date",
        "file_id",
        "D1",
        "D2",
        "fsc_small",
        "pe",
        "chl_small",
        "q2.5",
        "q50",
        "q97.5",
        "filter_id"
    ]
    df = df[columns]
    # Check for an existing file. Merge, overwriting matching existing entries.
    try:
        old_df = pd.read_parquet(outpath)
    except FileNotFoundError:
        pass
    else:
        if not all(old_df.columns == df.columns):
            raise ValueError("existing OPP parquet file has incompatible column names")
        new_files = list(df["file_id"].unique())
        old_df = old_df[~old_df["file_id"].isin(new_files)]  # drop rows in old_df that are in new data
        df = pd.concat([old_df, df], ignore_index=True)
        df.sort_values(by="file_id", kind="mergesort", inplace=True)  # mergesort is stable

    # Make sure file_id and filter_id are categorical columns
    if df["file_id"].dtype.name != "category":
        df["file_id"] = df["file_id"].astype("category")
    if df["filter_id"].dtype.name != "category":
        df["filter_id"] = df["filter_id"].astype("category")

    # Write parquet
    df.to_parquet(outpath, compression="snappy", index=False, engine="pyarrow")



import datetime
import logging
import os
import pathlib
import sys

import click
from seaflowpy import errors
from seaflowpy import seaflowfile
from seaflowpy import fileio
from seaflowpy import sample
from seaflowpy import sfl
from seaflowpy import time


def validate_file_fraction(ctx, param, value):
    if value <= 0 or value > 1:
        raise click.BadParameter('must be a number > 0 and <= 1.')
    return value


def validate_positive(ctx, param, value):
    if value is not None and value <= 0:
        raise click.BadParameter('must be a number > 0.')
    return value


def validate_seed(ctx, param, value):
    if value is not None:
        try:
            value = int(value)
        except ValueError as e:
            raise click.BadParameter('must be an integer: {}.'.format(e))
        if (value < 0 or value > (2**32 - 1)):
            raise click.BadParameter('must be between 0 and 2**32 - 1.')
    return value


def validate_timestamp(ctx, param, value):
    if value is not None:
        try:
            value = time.parse_date(value, assume_utc=False)
        except ValueError:
            raise click.BadParameter('unable to parse timestamp.')
    return value

def validate_hours(ctx, param, value):
    if value is not None:
        if value <= 0:
            raise click.BadParameter('hours must be > 0.')
    return value


@click.group()
def evt_cmd():
    """EVT file subcommand."""
    pass


@evt_cmd.command('sample')
@click.option('-o', '--outpath', type=click.Path(), required=True,
    help="""Output path for parquet file with subsampled event data.""")
@click.option('-c', '--count', type=int, default=100000, show_default=True, callback=validate_positive,
    help='Target number of events to keep.')
@click.option('-f', '--file-fraction', type=float, default=0.1, show_default=True, callback=validate_file_fraction,
    help='Fraction of files to sample from, > 0 and <= 1. Using --multi sets this option to 1.')
@click.option('--min-chl', type=int, default=0, show_default=True,
    help='Mininum chlorophyll (small) value.')
@click.option('--min-fsc', type=int, default=0, show_default=True,
    help='Mininum forward scatter (small) value.')
@click.option('--min-pe', type=int, default=0, show_default=True,
    help='Mininum phycoerythrin value.')
@click.option('--min-date', type=str, callback=validate_timestamp,
    help='Minimum date of file to sample as ISO8601 timestamp.')
@click.option('--max-date', type=str, callback=validate_timestamp,
    help='Maximum date of file to sample as ISO8601 timestamp.')
@click.option('--tail-hours', type=int, metavar='N', callback=validate_hours,
    help="""Only subsample the most recent N hours of data. Unsets --max-date.
            If --min-date is also provided it will be used if it is more recent
            than <last date - N hours>.""")
@click.option('--multi', is_flag=True, default=False, show_default=True,
    help='Sample --count events from each input file separately, rather than --count events overall.')
@click.option('-n', '--noise-filter', is_flag=True, default=False, show_default=True,
    help='Apply noise filter before subsampling.')
@click.option('-p', '--process-count', type=int, default=1, show_default=True, callback=validate_positive,
    help='Number of processes to use.')
@click.option('-s', '--seed', callback=validate_seed,
    help='Integer seed for PRNG, otherwise system-dependent source of randomness is used to seed the PRNG.')
@click.option('-S', '--sfl', 'sfl_path', type=click.Path(),
    help="""SFL file that can be used to associate dates with EVT files. Useful when
            sampling undated EVT files.""")
@click.option('-v', '--verbose', count=True,
    help='Show more information. Specify more than once to show more information.')
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def sample_evt_cmd(outpath, count, file_fraction, min_chl, min_fsc, min_pe,
                   min_date, max_date, tail_hours, multi, noise_filter, process_count, seed,
                   sfl_path, verbose, files):
    """
    Sample a subset of events in EVT files.

    The list of EVT files can be file paths or directory paths
    which will be searched for EVT files.
    COUNT events will be randomly selected from all data.
    """
    if verbose == 0:
        loglevel = logging.WARNING
    elif verbose == 1:
        loglevel = logging.INFO
    else:
        loglevel = logging.DEBUG
    logging.basicConfig(format="%(asctime)s:%(levelname)s:%(message)s", level=loglevel)

    # Get file to date mappings from SFL file
    files = seaflowfile.keep_evt_files(expand_file_list(files))
    files = seaflowfile.sorted_files(files)
    if sfl_path:
        sfl_df = sfl.read_file(sfl_path, convert_dates=True)
        sfl_df = sfl.fix(sfl_df)  # ensure valid file_ids in file column
    else:
        sfl_df = None
    evt = seaflowfile.date_evt_files(files, sfl_df)

    # Select by time.
    # If paths don't have timestamps and SFL not provided, this step will always
    # filter out all files.
    if tail_hours is not None and len(evt):
        tail_min_date = evt.date.max() - datetime.timedelta(hours=tail_hours)
        if min_date is None or min_date < tail_min_date:
            min_date = tail_min_date
        max_date = None
    if min_date is not None:
        evt = evt[evt.date >= min_date]
    if max_date is not None:
        evt = evt[evt.date <= max_date]

    # Select fraction of files
    if not multi:
        chosen_files = sample.random_select(list(evt.path), file_fraction, seed)
    else:
        chosen_files = list(evt.path)

    outdir = os.path.dirname(outpath)
    pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)

    results, errs = sample.sample(
        chosen_files,
        count,
        outpath,
        dates=dict(zip(list(evt.file_id), list(evt.date))),
        min_chl=min_chl,
        min_fsc=min_fsc,
        min_pe=min_pe,
        multi=multi,
        noise_filter=noise_filter,
        process_count=process_count,
        seed=seed
    )

    printed = False
    if verbose:
        if len(results):
            print("\t".join(["file_ID", "events", "postfilter_events", "sampled_events", "message"]), file=sys.stderr)
        for r in results:
            vals = [r["file_id"], r["events"], r["events_postfilter"], r["events_postsampling"], r["msg"]]
            print("\t".join([str(v) for v in vals]), file=sys.stderr)
        printed = True
    else:
        # Print only files that couldn't be read
        if len(results):
            print("\t".join(["file_ID", "events", "postfilter_events", "sampled_events", "message"]), file=sys.stderr)
        for r in results:
            if r["msg"]:
                print("\t".join([r["file_id"], r["msg"]]), file=sys.stderr)
                printed = True
    if printed:
        print("", file=sys.stderr)

    if errs:
        print("Errors encountered", file=sys.stderr)
        for err in errs:
            print(err, file=sys.stderr)
        print("", file=sys.stderr)

    if sfl_path:
        print("{} entries found in SFL file".format(len(sfl_df)), file=sys.stderr)
    print("{} input files".format(len(files)), file=sys.stderr)
    print("{} files within time window".format(len(evt)), file=sys.stderr)
    print("{} selected files".format(len(chosen_files)), file=sys.stderr)
    print("{} total events".format(sum([r["events"] for r in results])), file=sys.stderr)
    print("{} events after noise/min filtering".format(sum([r["events_postfilter"] for r in results])), file=sys.stderr)
    print("{} events sampled".format(sum([r["events_postsampling"] for r in results])), file=sys.stderr)



@evt_cmd.command('dates')
@click.option('--min-date', type=str, callback=validate_timestamp,
    help='Minimum date of file to consider.')
@click.option('--max-date', type=str, callback=validate_timestamp,
    help='Maximum date of file to consider.')
@click.option('--tail-hours', type=int, metavar='N', callback=validate_hours,
    help="""Only consider the most recent N hours of data. Unsets --max-date.
            If --min-date is also provided it will be used if it is more recent
            than <last date - N hours>.""")
@click.option('-S', '--sfl', 'sfl_path', type=click.Path(),
    help="""SFL file that can be used to associate dates with EVT files. Useful when
            sampling undated EVT files.""")
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def dates_evt_cmd(min_date, max_date, tail_hours, sfl_path, files):
    """
    Get date range for a set of EVT files.

    The list of EVT files can be file paths or directory paths
    which will be searched for EVT files.
    """
    logging.basicConfig(format="%(asctime)s:%(levelname)s:%(message)s", level=logging.INFO)
    files = seaflowfile.keep_evt_files(expand_file_list(files))
    files = seaflowfile.sorted_files(files)
    if sfl_path:
        sfl_df = sfl.read_file(sfl_path, convert_dates=True)
        sfl_df = sfl.fix(sfl_df)  # ensure valid file_ids in file column
    else:
        sfl_df = None
    evt = seaflowfile.date_evt_files(files, sfl_df)

    # Select by time.
    # If paths don't have timestamps and SFL not provided, this step will always
    # filter out all files.
    if tail_hours is not None and len(evt):
        tail_min_date = evt.date.max() - datetime.timedelta(hours=tail_hours)
        if min_date is None or min_date < tail_min_date:
            min_date = tail_min_date
        max_date = None
    if min_date is not None:
        evt = evt[evt.date >= min_date]
    if max_date is not None:
        evt = evt[evt.date <= max_date]

    if len(evt):
        print(f"{evt.date.min().isoformat()} {evt.date.max().isoformat()}")


@evt_cmd.command('validate')
@click.option('-a', '--all', 'report_all', is_flag=True,
    help='Show information for all files. If not specified then only files errors are printed.')
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def validate_evt_cmd(report_all, files):
    """
    Examines EVT files.

    If any of the file arguments are directories all EVT files within those
    directories will be recursively found and examined. Prints file validation
    report to STDOUT. Print summary of files passing validation to STDERR.
    """
    # TODO: expand to calculate hash for binary EVT
    # and OPP parquet files with an extra flag. OPP parquet should produce hashes for
    # each file_id in the whole file (e.g. by group_by)
    if not files:
        return

    # dirs to file paths
    files = expand_file_list(files)

    header_printed = False
    ok, bad = 0, 0

    for filepath in files:
        # Default values
        version = '-'
        file_id = '-'
        events = 0

        # Try to parse filename as SeaFlow file
        try:
            sff = seaflowfile.SeaFlowFile(filepath)
            file_id = sff.file_id
        except errors.FileError:
            # unusual name, no file_id
            pass

        # Try to read file as binary EVT
        try:
            data = fileio.read_evt_labview(filepath)
            version = data['version']
            status = 'OK'
            ok += 1
            events = len(data['df'].index)
        except errors.FileError as e:
            status = str(e)
            bad += 1

        if not header_printed:
            print('\t'.join(['path', 'file_id', 'version', 'status', 'events']))
            header_printed = True
        if (report_all and status == 'OK') or (status != 'OK'):
            print('\t'.join([filepath, file_id, version, status, str(events)]))
    print('%d/%d files passed validation' % (ok, bad + ok), file=sys.stderr)


def expand_file_list(files_and_dirs):
    """Convert directories in file list to EVT file paths."""
    # Find files in directories
    dirs = [f for f in files_and_dirs if os.path.isdir(f)]
    files = [f for f in files_and_dirs if os.path.isfile(f)]

    dfiles = []
    for d in dirs:
        dfiles = dfiles + seaflowfile.find_evt_files(d)

    return files + dfiles



"""Geo operations"""
import re


GGALAT_RE = re.compile(r'^(?P<degrees>-?\d{2})(?P<minutes>\d{2}(?:\.\d+)?)$')
GGALON_RE = re.compile(r'^(?P<degrees>-?\d{3})(?P<minutes>\d{2}(?:\.\d+)?)$')


def ggalat2dd(coord):
    """GGA latitude string (with +/- for N/S) to decimal degrees string.

    Precision to 4 decimal places (11.132 m)
    e.g. "2116.6922" -> "21.2782"
    """
    match = GGALAT_RE.match(coord)
    if not match:
        raise ValueError("Invalid GGA latitude string '{}'".format(coord))
    if match:
        degrees = int(match.group('degrees'))
        minutes = float(match.group('minutes'))
        if abs(degrees) > 90 or minutes > 60:
            raise ValueError("Invalid GGA latitude string '{}'".format(coord))
        sign = -1 if match.group('degrees')[0] == '-' else 1
    return "{:.4f}".format(sign * (abs(degrees) + (minutes / 60.0)))


def ggalon2dd(coord):
    """GGA longitude string (with +/- for E/W) to decimal degrees string.

    Precision to 4 decimal places (11.132 m)
    e.g. "2116.6922" -> "21.2782"
    """
    match = GGALON_RE.match(coord)
    if not match:
        raise ValueError("Invalid GGA longitude string '{}'".format(coord))
    if match:
        degrees = int(match.group('degrees'))
        minutes = float(match.group('minutes'))
        if abs(degrees) > 180 or minutes > 60:
            raise ValueError("Invalid GGA longitude string '{}'".format(coord))
        sign = -1 if match.group('degrees')[0] == '-' else 1
    return "{:.4f}".format(sign * (abs(degrees) + (minutes / 60.0)))


def is_gga_lat(coord):
    """Does this string look like a GGA latitude coordinate"""
    return bool(GGALAT_RE.match(coord))


def is_gga_lon(coord):
    """Does this string look like a GGA longitude coordinate"""
    return bool(GGALON_RE.match(coord))



import click
from seaflowpy.cli.commands.dayofyear_cmd import dayofyear_cmd
from seaflowpy.cli.commands.db_cmd import db_cmd
from seaflowpy.cli.commands.evt_cmd import evt_cmd
from seaflowpy.cli.commands.filter_cmd import filter_cmd
from seaflowpy.cli.commands.opp_cmd import opp_cmd
from seaflowpy.cli.commands.sfl_cmd import sfl_cmd
from seaflowpy.cli.commands.sds2sfl_cmd import sds2sfl_cmd
from seaflowpy.cli.commands.version_cmd import version_cmd

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    pass

cli.add_command(dayofyear_cmd, 'dayofyear')
cli.add_command(db_cmd, 'db')
cli.add_command(evt_cmd, 'evt')
cli.add_command(filter_cmd, 'filter')
cli.add_command(opp_cmd, 'opp')
cli.add_command(sds2sfl_cmd, 'sds2sfl')
cli.add_command(sfl_cmd, 'sfl')
cli.add_command(version_cmd, 'version')



import datetime
import glob
import os
import pathlib
import sys

import click
import pandas as pd

from seaflowpy import seaflowfile
from seaflowpy import time


def validate_positive(ctx, param, value):
    if value is not None and value <= 0:
        raise click.BadParameter('must be a number > 0.')
    return value


def validate_seed(ctx, param, value):
    if value is not None:
        try:
            value = int(value)
        except ValueError as e:
            raise click.BadParameter('must be an integer: {}.'.format(e))
        if (value < 0 or value > (2**32 - 1)):
            raise click.BadParameter('must be between 0 and 2**32 - 1.')
    return value


def validate_timestamp(ctx, param, value):
    if value is not None:
        try:
            value = time.parse_date(value, assume_utc=False)
        except ValueError:
            raise click.BadParameter('unable to parse timestamp.')
    return value

def validate_hours(ctx, param, value):
    if value is not None:
        if value <= 0:
            raise click.BadParameter('hours must be > 0.')
    return value


@click.group()
def opp_cmd():
    """OPP file subcommand."""
    pass


@opp_cmd.command('sample')
@click.option('-o', '--outpath', type=click.Path(), required=True,
    help="""Output path for parquet file with subsampled event data.""")
@click.option('-c', '--count', type=int, default=100000, show_default=True, callback=validate_positive,
    help='Target number of events to keep.')
@click.option('--min-date', type=str, callback=validate_timestamp,
    help='Minimum date of file to sample as ISO8601 timestamp.')
@click.option('--max-date', type=str, callback=validate_timestamp,
    help='Maximum date of file to sample as ISO8601 timestamp.')
@click.option('--tail-hours', type=int, metavar='N', callback=validate_hours,
    help="""Only subsample the most recent N hours of data. Unsets --max-date.
            If --min-date is also provided it will be used if it is more recent
            than <last date - N hours>.""")
@click.option('-s', '--seed', callback=validate_seed,
    help='Integer seed for PRNG, otherwise system-dependent source of randomness is used to seed the PRNG.')
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def sample_opp_cmd(outpath, count, min_date, max_date, tail_hours, seed,
                   files):
    """
    Sample a subset of events in OPP files.

    The list of OPP files can be file paths or directory paths
    which will be searched for OPP files.
    COUNT events will be randomly selected from all data.
    """
    files = sorted(expand_file_list(files))
    if files:
        timestamps = [os.path.basename(f).split(".")[0] for f in files]
        timestamps = [seaflowfile.timestamp_from_filename(ts) for ts in timestamps]
        timestamps = [datetime.datetime.fromisoformat(ts) for ts in timestamps]
        files_df = pd.DataFrame({"path": files, "date": timestamps})

        # Select by time.
        if tail_hours is not None and len(files_df):
            # Set min_date based on tail_hours from latest date
            opp = pd.read_parquet(files_df.path.values[-1], columns=["date"])
            tail_min_date = opp.date.max() - datetime.timedelta(hours=tail_hours)
            if min_date is None or min_date < tail_min_date:
                min_date = tail_min_date
            max_date = None

        if min_date is not None:
            files_df = files_df[files_df.date + datetime.timedelta(hours=1) >= min_date]
        if max_date is not None:
            files_df = files_df[files_df.date <= max_date]

        outdir = os.path.dirname(outpath)
        pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)

        df = pd.concat([pd.read_parquet(f) for f in files_df.path], ignore_index=True)
        if min_date is not None:
            df = df[df.date >= min_date]
        if max_date is not None:
            df = df[df.date <= max_date]
        count = min(count, len(df))
        if seed is not None:
            sub = df.sample(n=count, random_state=seed)
        else:
            sub = df.sample(n=count)
        sub.to_parquet(outpath)

        print("{} particles in time range".format(len(df)), file=sys.stderr)
        print("{} particles selected with time range {} - {} ".format(len(sub), sub.date.min().isoformat(), sub.date.max().isoformat()), file=sys.stderr)
    else:
        print("0 particles in time range", file=sys.stderr)


def expand_file_list(files_and_dirs):
    """Convert directories in file list to OPP file paths."""
    # Find files in directories
    dirs = [f for f in files_and_dirs if os.path.isdir(f)]
    files = [f for f in files_and_dirs if os.path.isfile(f)]
    dfiles = []
    for d in dirs:
        dfiles = dfiles + glob.glob(f"{d}/*.opp.parquet")
        dfiles = dfiles + glob.glob(f"{d}/**/*.opp.parquet")
    return files + dfiles



import pytest
from seaflowpy import time

# pylint: disable=redefined-outer-name

def test_parse_bad_date():
    with pytest.raises(ValueError):
        _ = time.parse_date("2019-11-06T10:32:a05+00:00")


def test_parse_good_date():
    dt = time.parse_date("2019-11-06T10:32:05+00:00")
    assert dt.isoformat() == "2019-11-06T10:32:05+00:00"


def test_parse_good_date_assume_utc():
    dt = time.parse_date("2019-11-06T10:32:05+07:00")
    # Should be parsed as UTC even though it's advertised otherwise
    assert dt.isoformat() == "2019-11-06T10:32:05+00:00"


def test_parse_good_date_notutc():
    dt = time.parse_date("2019-11-06T10:32:05+07:00", assume_utc=False)
    # Should not be parsed as UTC
    assert dt.isoformat() == "2019-11-06T10:32:05+07:00"


def test_parse_good_date_notz():
    dt = time.parse_date("2019-11-06T10:32:05", assume_utc=False)
    # Should be parsed as UTC
    assert dt.isoformat() == "2019-11-06T10:32:05+00:00"
    dt = time.parse_date("2019-11-06T10:32:05", assume_utc=True)
    # Should be parsed as UTC
    assert dt.isoformat() == "2019-11-06T10:32:05+00:00"



import click
from seaflowpy import seaflowfile
from seaflowpy import errors


@click.command()
@click.option('-v', '--verbose', is_flag=True,
    help='Print 3 columns: input path, file name, day of year dir.')
@click.argument('files', nargs=-1, type=click.Path())
def dayofyear_cmd(verbose, files):
    """
    Gets calculated day of year dir from filename timestamp.

    File paths must be new-style datestamped paths. Any part of the file
    path except for the filename will be ignored. The filename may include a
    '.gz' extension. Outputs to STDOUT.
    """
    output = []
    for file in files:
        try:
            sfile = seaflowfile.SeaFlowFile(file)
        except errors.FileError as e:
            click.echo("%s %s" % (file, e), err=True)
            continue
        if verbose:
            output.append([file, sfile.file_id, sfile.filename, sfile.dayofyear])
        else:
            output.append([sfile.dayofyear])
    if output:
        click.echo("\n".join(["\t".join(row) for row in output]))
